# coding: utf-8
from __future__ import annotations

import asyncio
import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.cluster.supervisor import FutureTomwerScanObserverWidget


class TestSupervisor(TestCaseQt):
    """Test the supervisor GUI"""

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()

        # set up scans
        self._scans = []
        self._future_tomo_objs = []
        for i in range(5):
            # create scan
            scan = MockNXtomo(
                scan_path=os.path.join(self.tempdir, f"scan_test{i}"),
                n_proj=10,
                n_ini_proj=10,
                create_ini_dark=False,
                create_ini_flat=False,
                dim=10,
            ).scan
            self._scans.append(scan)

            # create future
            future = asyncio.Future()
            if i == 1:
                future.set_result(None)
            self._future_tomo_objs.append(
                FutureTomwerObject(
                    tomo_obj=scan,
                    futures=(future,),
                )
            )
        # set up gui
        self._supervisor = FutureTomwerScanObserverWidget()
        self._supervisor.setConvertWhenFinished(False)

    def tearDown(self):
        self._supervisor.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._supervisor.close()
        self._supervisor = None
        shutil.rmtree(self.tempdir)
        self._scans.clear()
        self._future_tomo_objs.clear()
        super().tearDown()

    def test(self):
        for future_tomo_obj in self._future_tomo_objs:
            self._supervisor.addFutureTomoObj(future_tomo_obj=future_tomo_obj)
        self._supervisor.show()



"""Full tomwer test suite.

"""

from __future__ import annotations

import logging
import os
import shutil

from silx.gui import qt

from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.tests.datasets import TomwerCIDatasets

logger = logging.getLogger(__name__)


class Simulation(qt.QThread):
    """Simulation is a simple class able to simulate an acquisition by copying
    files on a targetted directory.

    :param targetdir: the folder where the acquisition is stored
    :param manipulationId: the id of the simulation we want to simulate
    :param finalState: when launched, the state to reach before stopping

    :warning: the targetted directory won't be removed or cleaned during class
        destruction. This is to be managed by callers.
    """

    advancement = {
        "not started": -1,
        "starting-s0": 0,
        "starting-s1": 1,
        "acquisitionRunning": 2,
        "acquisitionDone": 3,
        "reconstructionLaunched": 4,
    }
    sigAdvancementChanged = qt.Signal(int)

    __definedDataset = ["test01", "test10"]

    def __init__(self, targetdir, manipulationId, finalState=4):
        assert type(manipulationId) is str
        assert type(targetdir) is str
        assert manipulationId in self.__definedDataset
        super(Simulation, self).__init__()

        self.targetdir = targetdir
        self.outputFolder = os.path.sep.join((targetdir, manipulationId))
        self.finalState = finalState
        self.currentState = "not started"
        self._createFinalXML = False
        (
            self.originalFolder,
            self.nbSlices,
            self.manipulationId,
        ) = self.__getOriginalDataSet(manipulationId)
        self.stopFileCreationForRunningState = int(self.nbSlices / 2)
        self.srcPattern = None
        self.destPattern = None

    def __getOriginalDataSet(self, dataSetID):
        """Return paths to the requested scan"""
        assert dataSetID in self.__definedDataset
        dataDir = TomwerCIDatasets.get_dataset(os.path.join("edf_datasets", dataSetID))

        assert os.path.isdir(dataDir)
        assert os.path.isfile(os.path.join(dataDir, dataSetID + ".info"))
        slices = EDFTomoScan(dataDir).projections
        nbSlices = len(slices)
        manipulationID = dataSetID

        return dataDir, nbSlices, manipulationID

    def advanceTo(self, state):
        """Reset the new advancement targetted

        :param state: the new state to reach when run will be executed
        """
        assert state in Simulation.advancement
        assert type(state) is str
        self.finalState = Simulation.advancement[state]

    def setSrcDestPatterns(self, srcPattern, destPattern):
        """
        If setted, will set the .info and .xml files into a different folder
        """
        self.srcPattern = srcPattern
        self.destPattern = destPattern
        if srcPattern is not None or destPattern is not None:
            assert os.path.isdir(srcPattern)
            assert os.path.isdir(destPattern)
            targettedFolder = self.outputFolder.replace(
                self.srcPattern, self.destPattern, 1
            )
            if not os.path.isdir(targettedFolder):
                os.makedirs(targettedFolder)

    def __shouldExecStep(self, step):
        """Return True if the thread should exec this step to advance taking
        into consideration is current state and his final state
        """
        return self.finalState >= self.advancement[step] and (
            Simulation.advancement[self.currentState] + 1 == self.advancement[step]
        )

    def run(self):
        """Main function, run the acquisition through all states until
        finalState is reached
        """
        if self.__shouldExecStep("starting-s0") is True:
            logger.info("starting-s0")
            self._startAcquisition()
            self.currentState = "starting-s0"
            self.signalCurrentState()

        if self.__shouldExecStep("starting-s1") is True:
            self.copyInitialFiles()
            logger.info("starting-s1")
            self.currentState = "starting-s1"
            self.signalCurrentState()

        if self.__shouldExecStep("acquisitionRunning") is True:
            self._copyScans((0, self.stopFileCreationForRunningState))
            logger.info("acquisitionRunning")
            self.currentState = "acquisitionRunning"
            self.signalCurrentState()

        if self.__shouldExecStep("acquisitionDone") is True:
            self._copyScans((self.stopFileCreationForRunningState, self.nbSlices))
            if self._createFinalXML is True:
                inputXMLFile = os.path.join(
                    self.originalFolder, self.manipulationId + ".xml"
                )
                assert os.path.isfile(inputXMLFile)
                ouputXMLFile = os.path.join(
                    self.outputFolder, self.manipulationId + ".xml"
                )
                shutil.copyfile(inputXMLFile, ouputXMLFile)

            logger.info("acquisitionDone")
            self.currentState = "acquisitionDone"
            self.signalCurrentState()

    def signalCurrentState(self):
        """Signal the actual state of the simulation"""
        self.sigAdvancementChanged.emit(self.currentState)

    def _startAcquisition(self):
        """create needed data dir"""
        for newFolder in (self.targetdir, self.outputFolder):
            if not os.path.exists(self.outputFolder):
                os.makedirs(self.outputFolder)

    def _copyScans(self, _slicesRange):
        """copy the .edf file from the original directory to the outputFolder

        :_slicesRange tuple: the _range of slices data we want to copy
        """
        logger.info("copying files from %s to %s" % (_slicesRange[0], _slicesRange[1]))
        for iSlice in list(range(_slicesRange[0], _slicesRange[1])):
            filename = "".join((self.manipulationId, format(iSlice, "04d"), ".edf"))
            srcFile = os.path.join(self.originalFolder, filename)
            outputFile = os.path.join(self.outputFolder, filename)
            assert os.path.isfile(srcFile)
            assert os.path.isdir(self.outputFolder)
            shutil.copyfile(srcFile, outputFile)

    def copyInitialFiles(self):
        """copy the .info file"""
        assert os.path.isdir(self.originalFolder)
        logger.info(
            "copying initial files (.info, .xml...) from %s to %s"
            % (self.originalFolder, self.manipulationId)
        )
        for extension in (".info", ".db", ".cfg"):
            filename = "".join((self.manipulationId, extension))
            srcFile = os.path.join(self.originalFolder, filename)

            targettedFolder = self.outputFolder
            if self.srcPattern is not None or self.destPattern is not None:
                targettedFolder = self.outputFolder.replace(
                    self.srcPattern, self.destPattern, 1
                )
            assert os.path.isfile(srcFile)
            assert os.path.isdir(targettedFolder)
            assert os.path.isdir(self.originalFolder)
            shutil.copy2(srcFile, targettedFolder)

    def createFinalXML(self, val):
        """If activated, once all the file will be copied, this will create
        an .xml file into the output directory
        """
        self._createFinalXML = val

    def createParFile(self):
        pass

    def createReconstructedFile(self):
        pass

    def createOARJob(self):
        pass

    def createDark(self):
        pass

    def createJPG(self):
        pass

    def createVolfloat(self):
        pass

    def createVolraw(self):
        pass

    def __createFileTo(self, filePath):
        assert type(filePath) is str
        open(filePath, "a").close()



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.gui.stitching.axisorderedlist import EditableOrderedTomoObjWidget


def test_axis_ordered_list_widget(
    qtapp,  # noqa F811
):
    widget = EditableOrderedTomoObjWidget(axis=0)
    volumes = tuple(
        HDF5Volume(
            file_path=f"my_volume{i}.hdf5",
            data_path="my_volume",
        )
        for i in range(5)
    )
    [widget.addTomoObj(volume) for volume in volumes]

    widget.setSelectedTomoObjs([volumes[1], volumes[-1]])
    widget._callbackRemoveSelectedTomoObj()
    assert widget.getTomoObjsAxisOrdered() == (volumes[0], volumes[2], volumes[3])



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile
import unittest

import numpy

from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisTask
from tomwer.core.process.reconstruction.scores.scores import (
    _METHOD_TO_FCT,
    compute_score_contrast_std,
)
from tomwer.core.utils.scanutils import MockNXtomo


class TestScoreFunctions(unittest.TestCase):
    """Test all the score functions"""

    def test_img_contrast_std_score(self):
        """simple test of the API to call compute_score_contrast_std"""
        data = numpy.random.random(100 * 100).reshape(100, 100)
        compute_score_contrast_std(data)

    def test_method_to_function(self):
        """Test the dictionary used to for linking the score method to the
        callback function"""
        data = numpy.random.random(100 * 100).reshape(100, 100)
        for method_name, fct in _METHOD_TO_FCT.items():
            with self.subTest(method_name=method_name, fct=fct):
                res = fct(data)
                self.assertFalse(res is None)
                self.assertTrue(isinstance(res, float))


class TestSAAxisProcess(unittest.TestCase):
    """Test the SAAxisProcess class"""

    def setUp(self) -> None:
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=self.tempdir, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        self.scan = mock.scan

        self._default_saaxis_params = SAAxisParams()
        self._default_saaxis_params.output_dir = os.path.join(
            self.tempdir, "output_dir"
        )
        self._default_saaxis_params.slice_indexes = {"slice": 4}
        self._default_saaxis_params.nabu_config = {}
        self._default_saaxis_params.dry_run = True
        self._default_saaxis_params.file_format = "hdf5"

    def tearDown(self) -> None:
        shutil.rmtree(self.tempdir)

    def test(self):
        process = SAAxisTask(
            inputs={
                "data": self.scan,
                "sa_axis_params": self._default_saaxis_params.to_dict(),
                "serialize_output_data": False,
            }
        )

        self._default_saaxis_params.estimated_cor = 11
        self._default_saaxis_params.research_width = 2
        process = SAAxisTask(
            inputs={
                "data": self.scan,
                "sa_axis_params": self._default_saaxis_params.to_dict(),
                "serialize_output_data": False,
            },
        )
        process.run()
        process = SAAxisTask(
            inputs={
                "data": self.scan,
                "sa_axis_params": self._default_saaxis_params.to_dict(),
                "serialize_output_data": False,
            },
        )
        process.run()



from silx.utils.enum import Enum as _Enum


PROPOSAL_GALLERY_DIR_NAME = "GALLERY"
DATASET_GALLERY_DIR_NAME = "gallery"


class OutputFormat(_Enum):
    """possible output format to save screenshots"""

    PNG = "png"
    JPEG = "jpg"



from __future__ import annotations

from silx.gui import qt
from tomwer.core.process.drac.output import OutputFormat
from tomwer.gui import icons

from tomwer.core.process.drac.binning import Binning


class GalleryOptionsAction(qt.QAction):
    def __init__(self, parent):
        icon = icons.getQIcon("icat_gallery_opts")
        qt.QAction.__init__(self, icon, "Icat gallery options", parent)


class GalleryWidget(qt.QWidget):
    """Widget to let the user define the output location of the screenshots"""

    sigConfigChanged = qt.Signal()
    """emit when the configuration has changed"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())

        # screenshot precision
        self._precisionQCB = qt.QComboBox(self)
        self._precisionQCB.addItem("uint8")
        self.layout().addRow("precision", self._precisionQCB)
        # binning
        self._binningQCB = qt.QComboBox(self)
        self._binningQCB.addItems(Binning.values())
        self.layout().addRow("binning", self._binningQCB)
        self._binningQCB.setCurrentText(Binning.SIXTEEN_BY_SIXTEEN.value)
        self._binningQCB.setToolTip(
            "To speed up display of the gallery at the data portal side it is highly recommended to bin screenshots"
        )  # recommended size: 5ko for the entire gallery
        # output format
        self._outputFormat = qt.QComboBox(self)
        self._outputFormat.addItems(OutputFormat.values())
        self.layout().addRow("output format", self._outputFormat)
        # overwrite
        self._overwriteCB = qt.QCheckBox("overwrite", self)
        self._overwriteCB.setChecked(True)
        self.layout().addRow(self._overwriteCB)
        # connect signal / slot
        self._outputFormat.currentIndexChanged.connect(self._configChanged)
        self._overwriteCB.toggled.connect(self._configChanged)
        self._binningQCB.currentIndexChanged.connect(self._configChanged)

    def getOutputFormat(self) -> OutputFormat:
        return OutputFormat.from_value(self._outputFormat.currentText())

    def setOutputFormat(self, format: OutputFormat):
        format = OutputFormat.from_value(format)
        self._outputFormat.setCurrentText(format.value)

    def getBinning(self) -> Binning:
        return Binning.from_value(self._binningQCB.currentText())

    def setBinning(self, binning: Binning):
        binning = Binning.from_value(binning)
        self._binningQCB.setCurrentText(binning.value)

    def overwrite(self) -> bool:
        return self._overwriteCB.isChecked()

    def setOverwrite(self, overwrite: bool) -> None:
        return self._overwriteCB.setChecked(overwrite)

    def getConfiguration(self):
        return {
            "output_format": self.getOutputFormat().value,
            "overwrite": self.overwrite(),
            "binning": self.getBinning().value,
        }

    def setConfiguration(self, config: dict):
        if not isinstance(config, dict):
            raise TypeError(f"config is a expected to be a dict. got {type(config)}")
        output_format = config.get("output_format", None)
        if output_format is not None:
            self.setOutputFormat(output_format)

        overwrite = config.get("overwrite", None)
        if overwrite is not None:
            overwrite = overwrite in (True, "True", 1)
            self.setOverwrite(overwrite=overwrite)

        binning = config.get("binning", None)
        if binning is not None:
            self.setBinning(binning=binning)

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()


class GalleryDialog(qt.QDialog):
    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._widget = GalleryWidget()
        self.layout().addWidget(self._widget)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

    # expose API
    def setConfiguration(self, gallery_options: dict) -> None:
        self._widget.setConfiguration(gallery_options)

    def getConfiguration(self) -> dict:
        return self._widget.getConfiguration()

    def getOutputFormat(self) -> OutputFormat:
        return self._widget.getOutputFormat()

    def setOutputFormat(self, format: OutputFormat):
        return self._widget.setOutputFormat(format=format)

    def getBinning(self) -> Binning:
        return self._widget.getBinning()

    def setBinning(self, binning: Binning):
        self._widget.setBinning(binning=binning)

    def overwrite(self) -> bool:
        return self._widget.overwrite

    def setOverwrite(self, overwrite: bool) -> None:
        self._widget.setOverwrite(overwrite=overwrite)



"""
contains gui relative to axis calculation using sinogram
"""

from __future__ import annotations

import logging

from silx.gui import qt
from tomwer.gui.utils.qt_utils import block_signals

from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.process.reconstruction.utils.cor import (
    absolute_pos_to_relative,
    relative_pos_to_absolute,
)
from tomwer.core.scan.scanbase import TomwerScanBase

from ...utils.scandescription import ScanNameLabelAndShape
from .AxisWidget import AxisWidget
from .ControlWidget import ControlWidget

_logger = logging.getLogger(__file__)


class AxisMainWindow(qt.QMainWindow):
    """
    Main window for the center of rotation search.
    It displays:
    * scan information
    * AxisWidget
    * allows to set the cor value as absolute and relative and edit it
    """

    sigComputationRequested = qt.Signal()
    """emit when the user request for computation"""

    sigApply = qt.Signal()
    """emit when the user validate the axis value"""

    sigLockCORValueChanged = qt.Signal(bool)
    """bool: True if locked"""

    sigLockModeChanged = qt.Signal()
    """Signal emitted when the lock mode on the mode change"""

    sigModeChanged = qt.Signal(str)
    """signal emit when the mode is changed"""

    sigSinogramReady = qt.Signal()
    """signal emit when the sinogram load is complete and processing ready"""

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    def __init__(self, axis_params, parent=None, backend=None):
        super().__init__(parent)

        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QVBoxLayout())

        # add scan name
        self._scan_label = ScanNameLabelAndShape(parent=self)
        self._mainWidget.layout().addWidget(self._scan_label)

        self.setDockOptions(qt.QMainWindow.AnimatedDocks)

        # Axis Widget
        self._axis_params = axis_params
        self._axisWidget = AxisWidget(
            parent=self, axis_params=axis_params, backend=backend
        )
        self._mainWidget.layout().addWidget(self._axisWidget)
        self.setCentralWidget(self._mainWidget)

        # append cor editor to the radio axis widget
        self._controlWidget = ControlWidget(parent=self)
        self._controlWidget.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self._controlDockWidget = qt.QDockWidget(parent=self)
        self._controlDockWidget.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self._controlDockWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._controlDockWidget.layout().setSpacing(0)
        self._controlDockWidget.setMaximumHeight(150)
        self._controlDockWidget.setWidget(self._controlWidget)
        self._controlDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._axisWidget.addDockWidget(
            qt.Qt.RightDockWidgetArea, self._controlDockWidget
        )

        # set up
        if self._axis_params.mode is AxisMode.manual:
            self._controlWidget.setPosition(
                self._axis_params.relative_cor_value,
                self._axis_params.absolute_cor_value,
            )
        self._controlWidget._positionInfo.setAxisParams(self._axis_params)

        # connect signal / slots
        self._controlWidget.sigComputationRequest.connect(self.sigComputationRequested)
        self._controlWidget.sigValidateRequest.connect(self.sigApply)
        self._controlWidget.sigLockCORValueChanged.connect(self._CORValueLocked)
        self._axisWidget.sigLockModeChanged.connect(self.sigLockModeChanged)
        self._axisWidget.sigPositionChanged.connect(self._setPositionFrmTuple)
        self._controlWidget._positionInfo.sigRelativeValueSet.connect(
            self._forceRelativePosition
        )
        self._controlWidget._positionInfo.sigAbsolueValueSet.connect(
            self._forceAbsolutePosition
        )

    def _forceRelativePosition(self, value: float):
        self.setMode(AxisMode.manual)
        self.setPosition(
            relative_value=value,
            absolute_value=None,
        )

    def _forceAbsolutePosition(self, value: float):
        self.setMode(AxisMode.manual)
        self.setPosition(
            relative_value=None,
            absolute_value=value,
        )

    def setAutoUpdateEstimatedCor(self, value):
        self._axisWidget._settingsWidget._mainWidget.setUpdateXRotationAxisPixelPositionOnNewScan(
            value
        )

    def manual_uses_full_image(self, value):
        self._axisWidget.manual_uses_full_image(value)

    def _modeChanged(self):
        self.getAxisParams().mode = self.getMode()

    def _CORValueLocked(self, lock):
        if lock:
            self.setMode(AxisMode.manual)
        self.setModeLock(lock)
        self.sigLockCORValueChanged.emit(lock)

    def _setPositionFrmTuple(self, value):
        self.setPosition(relative_value=value[0])

    def setPosition(self, relative_value: float, absolute_value: float = None) -> None:
        if (
            absolute_value is None
            and self._axis_params.frame_width is not None
            and relative_value is not None
        ):
            try:
                absolute_value = relative_pos_to_absolute(
                    relative_pos=relative_value, det_width=self._axis_params.frame_width
                )
            except TypeError:
                absolute_value = None
        if (
            relative_value is None
            and self._axis_params.frame_width is not None
            and absolute_value is not None
        ):
            try:
                relative_value = absolute_pos_to_relative(
                    absolute_pos=absolute_value, det_width=self._axis_params.frame_width
                )
            except TypeError:
                relative_value = None
        self._controlWidget.setPosition(
            relative_cor=relative_value, abs_cor=absolute_value
        )
        if relative_value is not None:
            self._axisWidget.setXShift(relative_value)
        else:
            self._axisWidget.resetShift()

    def getAxisParams(self):
        return self._axis_params

    def setScan(self, scan: TomwerScanBase, set_position: bool = False):
        """
        set the gui for this scan

        :param scan:
        """
        self._axisWidget.setScan(scan=scan)
        if set_position is True and scan.axis_params is not None:
            self.setPosition(
                scan.axis_params.relative_cor_value, scan.axis_params.relative_cor_value
            )
        elif set_position is False:
            if scan.dim_1 is not None:
                # if absolute position is know and relative position not (in the case the user already define absolute position only):
                has_absolute_val_already = not self._controlWidget._positionInfo._absolutePositionQLE.text().startswith(
                    (".", "?")
                )
                has_relative_val_already = not self._controlWidget._positionInfo._relativePositionQLE.text().startswith(
                    (".", "?")
                )
                if (
                    self.getMode() is AxisMode.manual
                    and has_absolute_val_already
                    and not has_relative_val_already
                ):
                    self._controlWidget._updateRelativePosition(width=scan.dim_1)
                else:
                    self._controlWidget._updateAbsolutePosition(width=scan.dim_1)

    def _computationRequested(self) -> None:
        self.sigComputationRequested.emit()

    def _setModeLockFrmSettings(self, lock):
        with block_signals(self):
            self._axisWidget._setModeLockFrmSettings(lock)

    def _setValueLockFrmSettings(self, lock):
        with block_signals(self):
            self.setValueLock(lock)

    def setModeLock(self, lock):
        assert type(lock) is bool
        self._axisWidget.setLocked(lock)

    def isModeLock(self):
        return self._axisWidget.isModeLock()

    def isValueLock(self):
        return self._controlWidget.isValueLock()

    def setValueLock(self, lock):
        self._controlWidget.setValueLock(lock)

    def setReconsParams(self, recons_params):
        self._axis_params = recons_params
        self._axisWidget.setReconsParams(axis=recons_params)
        if recons_params.mode is AxisMode.manual:
            self._controlWidget.setPosition(
                self._axis_params.relative_cor_value,
                self._axis_params.absolute_cor_value,
            )

    # expose API
    def hideLockButton(self) -> None:
        self._controlWidget.hideLockButton()

    def hideApplyButton(self) -> None:
        self._controlWidget.hideApplyButton()

    def setMode(self, mode):
        mode = AxisMode.from_value(mode)
        self._axisWidget.setMode(mode)
        self._axis_params.mode = mode
        self.sigModeChanged.emit(mode.value)

    def getMode(self):
        return self._axisWidget.getMode()

    def getEstimatedCor(self) -> float:
        return self._axisWidget.getEstimatedCor()

    def setEstimatedCor(self, value: float):
        self._axisWidget.setEstimatedCor(value=value)

    def getAutoUpdateEstimatedCor(self):
        return self._axisWidget.updateXRotationAxisPixelPositionOnNewScan()

    def isYAxisInverted(self) -> bool:
        return self._axisWidget.isYAxisInverted()

    def setYAxisInverted(self, checked: bool):
        return self._axisWidget.setYAxisInverted(checked=checked)



import os
import pickle

import numpy
import pytest
from nxtomomill.nexus.nxtomo import NXtomo
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui.utils.testutils import SignalListener
from nxtomo.nxobject.nxdetector import ImageKey

from orangecontrib.tomwer.widgets.edit.NXtomoEditorOW import NXtomoEditorOW
from tomwer.core.process.edit.nxtomoeditor import NXtomoEditorKeys
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.tests.utils import skip_gui_test
from tomwer.gui.utils.qt_utils import QSignalSpy
from tomwer.tests.conftest import qtapp  # noqa F401


def getDefaultConfig() -> dict:
    """return the configuration of the NXtomo editor. First value is the value of the field, second is: is the associated lock button locked or not"""
    return {
        NXtomoEditorKeys.ENERGY: (5.9, False),
        NXtomoEditorKeys.SAMPLE_DETECTOR_DISTANCE: (2.4, True),
        NXtomoEditorKeys.FIELD_OF_VIEW: ("Full", False),
        NXtomoEditorKeys.X_PIXEL_SIZE: (0.023, True),
        NXtomoEditorKeys.Y_PIXEL_SIZE: (0.025, True),
        NXtomoEditorKeys.X_FLIPPED: (True, True),
        NXtomoEditorKeys.Y_FLIPPED: (False, False),
        NXtomoEditorKeys.X_TRANSLATION: (0.0,),
        NXtomoEditorKeys.Z_TRANSLATION: (0.0,),
    }


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_NXtomoEditorOW(
    qtapp,  # noqa F811
    tmp_path,
):
    window = NXtomoEditorOW()
    # test serialization
    window.setConfiguration(getDefaultConfig())
    assert window.getConfiguration() == getDefaultConfig()
    pickle.dumps(window.getConfiguration())

    # test literal dumps
    literal_dumps(window.getConfiguration())

    # test widget automation
    signal_listener = SignalListener()
    window.sigScanReady.connect(signal_listener)
    # set up the widget to define and lock distance, energy and x pixel size
    distance_widget = window.widget.mainWidget._distanceMetricEntry
    distance_widget.setValue(0.6)
    distance_widget.setUnit("mm")
    distance_locker = window.widget.mainWidget._distanceLB
    distance_locker.setLock(True)
    energy_widget = window.widget.mainWidget._energyEntry
    energy_widget.setValue(88.058)
    energy_locker = window.widget.mainWidget._energyLockerLB
    energy_locker.setLock(True)
    x_pixel_widget = window.widget.mainWidget._xPixelSizeMetricEntry
    x_pixel_widget.setValue(45)
    x_pixel_widget.setUnit("nm")
    x_pixel_locker = window.widget.mainWidget._xPixelSizeLB
    x_pixel_locker.setLock(True)

    # 1.0 create nx tomos with raw data
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.x_pixel_size = (
        0.023  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.y_pixel_size = (
        0.025  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.field_of_view = "full"
    nx_tomo.instrument.detector.distance = (
        2.4  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.x_flipped = (
        False  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.y_flipped = True
    nx_tomo.energy = 5.9
    nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo.sample.rotation_angle = numpy.linspace(0, 20, num=12)

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo.save(
        file_path=file_path,
        data_path=entry,
    )
    # 2.0 set scan to the nxtomo-editor
    scan = NXtomoScan(file_path, entry)
    waiter = QSignalSpy(window.sigScanReady)
    window.setScan(scan=scan)
    # warning: avoid executing the ewoks task as this will be done
    # automatically (has some field lock). This would create concurrency task
    # and could bring some HDF5 concurrency error
    waiter.wait(5000)
    # 3.0 check results are as expected
    # make sure the scan has been re-emitted
    assert signal_listener.callCount() == 1
    # make sure the edition of the parameters have been done and only those
    overwrite_nx_tomo = NXtomo().load(
        file_path=file_path,
        data_path=entry,
        detector_data_as="as_numpy_array",
    )
    numpy.testing.assert_almost_equal(
        overwrite_nx_tomo.instrument.detector.x_pixel_size.si_value, 45e-9
    )
    assert (
        overwrite_nx_tomo.instrument.detector.y_pixel_size.si_value
        == nx_tomo.instrument.detector.y_pixel_size.si_value
    )
    assert (
        overwrite_nx_tomo.instrument.detector.field_of_view
        == nx_tomo.instrument.detector.field_of_view
    )
    numpy.testing.assert_almost_equal(
        overwrite_nx_tomo.instrument.detector.distance.value, 6.0e-4
    )
    assert (
        overwrite_nx_tomo.instrument.detector.x_flipped
        == True  # is lock to True so should load the value and overwrite it
    )

    assert (
        overwrite_nx_tomo.instrument.detector.y_flipped
        == True  # is not locked and is initially set to False (should load it and ignore overwriting)
    )
    numpy.testing.assert_almost_equal(overwrite_nx_tomo.energy.value, 88.058)
    numpy.testing.assert_array_almost_equal(
        overwrite_nx_tomo.instrument.detector.data,
        nx_tomo.instrument.detector.data,
    )



"""
Define the tomwer Task class.
Insure connection with ewoks.
All instances of tomwer Tasks should avoid gui import (as qt).
"""

from __future__ import annotations

import logging
from collections import namedtuple
from datetime import datetime

import h5py
import numpy
from ewokscore.task import Task as _EwoksTask
from ewokscore.taskwithprogress import TaskWithProgress as _EwoksTaskWithProgress
from silx.io.dictdump import dicttoh5, h5todict
from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5
from tomoscan.io import HDF5File
from tomwer.core.utils.locker import FileLockerManager


_logger = logging.getLogger(__name__)

_process_desc = namedtuple(
    "_process_desc", ["process_order", "configuration", "results"]
)


class BaseProcessInfo:
    """Tomwer base process class"""

    _output_values = {}
    # TODO: look at this: not sure this is needed anymore

    def __init__(self, inputs=None):
        """
        :param return_dict: if True serialize (to_dict / from_dict functions) between each task
        """
        if inputs is None:
            inputs = {}

        self._scheme_title = (
            "scheme_title"  # TODO: have a look, this must be get somewhere and reused ?
        )

        """should the return type of the handler should be TomoBase instance
        objects or dict"""
        self._settings = {}
        self._cancelled = False
        # a useful variable that can be set to True if the task has been cancelled

    @staticmethod
    def properties_help():
        """

        :return: display the list of all managed keys and possible values
        """
        # TODO: use argsparse instead of this dict ?
        raise NotImplementedError("BaseProcess is an abstract class")

    def get_output_value(self, key):
        """

        :param key:
        :return:
        """
        assert type(key) is str
        if key in self._output_values:
            return self._output_values[key]
        else:
            return None

    def clear_output_values(self):
        self._output_values.clear()

    def register_output(self, key, value):
        """

        :param key: name of the output
        :param value: value of the output
        """
        self._output_values[key] = value

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        raise NotImplementedError("Base class")

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        raise NotImplementedError("Base class")

    @staticmethod
    def definition():
        """definition of the process"""
        raise NotImplementedError("Base class")

    def get_configuration(self) -> dict | None:
        """

        :return: configuration of the process
        """
        if self._settings is None:
            return None
        if len(self._settings) > 0:
            return self._settings
        else:
            return None

    def set_configuration(self, configuration: dict) -> None:
        self._settings = configuration

    def register_process(
        self,
        process_file: str,
        entry: str,
        configuration: dict | None,
        results: dict | None,
        process_index: int,
        interpretations: dict | None = None,
        overwrite: bool = True,
    ) -> None:
        """
        Store the current process in the linked h5 file if any,
        output data stored will be the one defined by the data_keys

        :param process_file: where to store the processing information
        :param entry: entry process
        :param configuration: configuration of the process
        :param results: result of the processing
        :param process_index: index of the process
        :param overwrite: if True then overwrite the process if already
                               exists
        :paramerpretations: for each result we can add a flag
                                'interpretation'
        """
        assert process_file is not None
        with FileLockerManager().get_lock(file_name=process_file):
            try:
                self._register_process(
                    process_file=process_file,
                    entry=entry,
                    process=self,
                    configuration=configuration,
                    results=results,
                    process_index=process_index,
                    interpretations=interpretations,
                    overwrite=overwrite,
                )
            except IOError as e:
                _logger.error(e)

    @staticmethod
    def _register_process(
        process_file: str,
        process,
        entry: str | None,
        configuration: dict | None,
        results: dict | None,
        process_index: int,
        interpretations: dict | None = None,
        overwrite: bool = True,
    ) -> None:
        """
        Store the current process in the linked h5 file if any,
        output data stored will be the one defined by the data_keys

        :param process: process to record
        :param process_file: where to store the processing information
        :param entry: entry process
        :param configuration: configuration of the process
        :param results: result of the processing
        :param process_index: index of the process
        :param overwrite: if True then overwrite the process if already
                               exists
        :paramerpretations: for each result we can add a flag
                                'interpretation'
        """
        assert process_file is not None, "The process file should be defined"
        if interpretations is None:
            interpretations = {}
        process_name = "tomwer_process_" + str(process_index)

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file (lock should be handled upper)
        with HDF5File(process_file, mode="a") as h5f:
            nx_process = h5f.require_group(get_process_path())
            if "NX_class" not in nx_process.attrs:
                nx_process.attrs["NX_class"] = "NXprocess"
            if overwrite:
                for key in (
                    "program",
                    "results",
                    "version",
                    "date",
                    "sequence_index",
                    "class_instance",
                    "configuration",
                ):
                    if key in nx_process:
                        del nx_process[key]

            # write process information
            nx_process["program"] = process.program_name()
            nx_process["version"] = process.program_version()
            nx_process["date"] = datetime.now().replace(microsecond=0).isoformat()
            nx_process["sequence_index"] = numpy.int32(process_index)
            _class = process.__class__
            nx_process["class_instance"] = ".".join(
                (_class.__module__, _class.__name__)
            )

        # dump result
        if results is not None:
            h5path = "/".join((get_process_path(), "results"))
            dicttoh5(
                results,
                h5file=process_file,
                h5path=h5path,
                update_mode="modify",
                mode="a",
            )
            for interpretation_key, interpretation in interpretations.items():
                try:
                    with HDF5File(process_file, mode="a") as h5f:
                        node = h5f["/".join((get_process_path(), "results"))]
                        node[interpretation_key].attrs[
                            "interpretation"
                        ] = interpretation
                except KeyError:
                    _logger.warning(
                        "Invalid interpretation - no result store"
                        " for %s" % interpretation_key
                    )

        # dump configuration
        if configuration is not None:
            h5path = "/".join((get_process_path(), "configuration"))
            dicttoh5(
                configuration,
                h5file=process_file,
                h5path=h5path,
                update_mode="modify",
                mode="a",
            )

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                nx_process["configuration"].attrs["NX_class"] = "NXcollection"

    @staticmethod
    def _get_process_nodes(
        root_node: h5py.Group, process, version: str = None, depth: int = 2
    ) -> dict:
        """

        :param root_node:
        :param version: version to look for (ignored for now)
        :return: list of process nodes [as h5py.Group]. Key is node name,
                 value is the process index
        """

        def is_process_node(node):
            if node is None:
                return False
            return (
                node.name.split("/")[-1].startswith("tomwer_process_")
                and "NX_class" in node.attrs
                and node.attrs["NX_class"] == "NXprocess"
                and "program" in node
                and (
                    process is None
                    or h5py_read_dataset(node["program"]) == process.program_name()
                )
                and "version" in node
                and "sequence_index" in node
            )

        if isinstance(root_node, h5py.Dataset):
            return {}
        res = {}
        if is_process_node(root_node):
            res[root_node.name] = int(h5py_read_dataset(root_node["sequence_index"]))

        if root_node is not None:
            for _, node in root_node.items():
                if depth >= 1:
                    res.update(
                        Task._get_process_nodes(
                            process=process,
                            root_node=node,
                            depth=depth,
                            version=version,
                        )
                    )
        return res

    @staticmethod
    def get_most_recent_process(
        root_node: h5py.Group, process, version: str = None
    ) -> h5py.Group | None:
        nodes = Task._get_process_nodes(
            root_node=root_node, process=process, version=version, depth=1
        )
        nodes_with_time = []
        nodes_time = []
        import datetime

        for node_name in nodes:
            node = root_node[node_name]
            assert isinstance(node, h5py.Group), "node should be a h5py.Group"
            if "date" in node:
                nodes_with_time.append(node)
                nodes_time.append(
                    datetime.datetime.fromisoformat(h5py_read_dataset(node["date"]))
                )
        if len(nodes_with_time) == 0:
            return None
        else:
            st = numpy.argmax(nodes_time)
            return nodes_with_time[st]

    @staticmethod
    def get_processes_frm_type(process_file: str, process_type, entry=None) -> list:
        """

        :param process_file: file to read
        :param process_type: process type we are looking for
        :return: list of _process_desc(sequence_index, configuration, results)
        """
        # retrieve process to load
        with open_hdf5(process_file) as h5f:
            if entry is None:
                if len(h5f.keys()) > 0:
                    root = h5f[list(h5f.keys())[0]]
                else:
                    _logger.warning("no process find")
                    return []
            else:
                root = h5f[entry]

            processes_to_load = {}
            for process in root.keys():
                nx_process = root[process]
                if (
                    h5py_read_dataset(nx_process["program"])
                    == process_type.program_name()
                ):
                    processes_to_load[nx_process.name] = h5py_read_dataset(
                        nx_process["sequence_index"]
                    )

        # load process
        processes = []
        for process_name, p_order in processes_to_load.items():
            config = h5todict(process_file, "/".join((process_name, "configuration")))
            results = h5todict(process_file, "/".join((process_name, "results")))
            processes.append(
                _process_desc(
                    process_order=p_order, configuration=config, results=results
                )
            )
        return processes


class TaskWithProgress(_EwoksTaskWithProgress, BaseProcessInfo):
    """Class from which all tomwer process should inherit

    :param logger: the logger used by the class
    """

    def __init__(
        self,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
        progress=None,
    ):
        BaseProcessInfo.__init__(self, inputs=inputs)
        _EwoksTaskWithProgress.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
            progress=progress,
        )


class Task(_EwoksTask, BaseProcessInfo):
    """Class from which all tomwer process should inherit

    :param logger: the logger used by the class
    """

    def __init__(
        self,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        BaseProcessInfo.__init__(self, inputs=inputs)
        _EwoksTask.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )



"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

from __future__ import annotations

import logging
import os
import h5py
import tempfile

from silx.io.dictdump import dicttoh5, h5todict
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5
from tomoscan.esrf.scan.utils import (
    copy_h5_dict_darks_to,
    copy_h5_dict_flats_to,
    cwd_context,
)
from tomoscan.framereducer.target import REDUCER_TARGET
from tomoscan.io import HDF5File

from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefsTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import data_identifier_to_scan

logger = logging.getLogger(__name__)


class DarkRefsCopy(DarkRefsTask):
    """
    Reimplement Dark ref to deal with copy when there is no median/mean files
    """

    DEFAULT_SRCURRENT = 200.0  # mA

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        super().__init__(
            process_id=process_id,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        if inputs is None:
            inputs = {}

        self._mode_auto = inputs.get("mode_auto", True)
        if "savedir" in inputs:
            raise KeyError("savedir is not a valid key. Use save_dir")
        self._savedir = inputs.get("save_dir", tempfile.mkdtemp())
        self._save_file = self.get_save_file(self._savedir)
        self._darks_url = self.get_darks_url(self._savedir)
        self._flats_url = self.get_flats_url(self._savedir)
        logger.info(f"Flats and darks will be stored in {self._save_file}")

        self._processOnlyCopy = inputs.get("process_only_copy", False)
        self._processOnlyDkRf = inputs.get("process_only_dkrf", False)
        init_ref_scan = inputs.get("init_dkrf_from", None)
        if init_ref_scan:
            if not isinstance(init_ref_scan, TomwerScanBase):
                raise TypeError(
                    f"init_ref_scan is expected to be an instance of TomwerScanBase. Not {type(init_ref_scan)}"
                )
            else:
                self.set_darks_and_flats_from_scan(init_ref_scan)

    @staticmethod
    def get_save_file(save_dir):
        return os.path.join(save_dir, "darks_and_flats.h5")

    @staticmethod
    def get_flats_url(save_dir):
        return DataUrl(
            file_path=DarkRefsCopy.get_save_file(save_dir),
            data_path="/flats",
            scheme="silx",
        )

    @staticmethod
    def get_darks_url(save_dir):
        return DataUrl(
            file_path=DarkRefsCopy.get_save_file(save_dir),
            data_path="/darks",
            scheme="silx",
        )

    def set_process_only_dkRf(self, value: bool) -> None:
        self._processOnlyDkRf = value
        self._processOnlyCopy = False

    def set_process_only_copy(self, value: bool) -> None:
        self._processOnlyDkRf = False
        self._processOnlyCopy = value

    @staticmethod
    def get_reduced_frame_data(
        url: DataUrl, reduced_target: REDUCER_TARGET, check_for_reduced_key: bool = True
    ) -> tuple:
        """
        :param url: url expected to contain the dict with one key per serie of reduced frame
        :param reduced_target: what we are looking for (used when check_for_reduced_key is set to True)
        :param check_for_reduced_key: if True then if the dict contains the given key then return the value of this key instead
                                           of the loaded dict. Used as a user friendly feature to be able to load dict when upper data path
                                           is provided instead of the 'darks' or 'flats' one.
        :return: tuple as (data, metadata)
        """
        reduced_target = REDUCER_TARGET.from_value(reduced_target)

        with cwd_context(url.file_path()):
            try:
                reduced_info_dict = h5todict(
                    h5file=url.file_path(),
                    path=url.data_path(),
                )
            except Exception as e:
                logger.error(e)
                return None
            else:
                if check_for_reduced_key and reduced_target.value in reduced_info_dict:
                    return reduced_info_dict[reduced_target.value]
                else:
                    return reduced_info_dict

    @staticmethod
    def save_flats_to_be_copied(save_dir, data: DataUrl | dict):
        if isinstance(data, DataUrl):
            data = DarkRefsCopy.get_reduced_frame_data(url=data, reduced_target="flats")
        if data is None:
            return
        flat_url = DarkRefsCopy.get_flats_url(save_dir)
        # remove group if already exists: else can end up by copying several dark / flat from different datasets
        with HDF5File(flat_url.file_path(), mode="a") as h5f:
            if flat_url.data_path() in h5f:
                del h5f[flat_url.data_path()]
        dicttoh5(
            data,
            h5file=flat_url.file_path(),
            h5path=flat_url.data_path(),
            mode="a",
            update_mode="replace",
        )

    @staticmethod
    def save_darks_to_be_copied(save_dir, data: DataUrl | dict):
        if isinstance(data, DataUrl):
            data = DarkRefsCopy.get_reduced_frame_data(url=data, reduced_target="darks")
        if data is None:
            return
        dark_url = DarkRefsCopy.get_darks_url(save_dir)
        # remove group if already exists: else can end up by copying several dark / flat from different datasets
        with HDF5File(dark_url.file_path(), mode="a") as h5f:
            if dark_url.data_path() in h5f:
                del h5f[dark_url.data_path()]
        dicttoh5(
            data,
            h5file=dark_url.file_path(),
            h5path=dark_url.data_path(),
            mode="a",
            update_mode="replace",
        )

    def clear_cache(self):
        """
        remove the file used to cache the reduced darks / flats.
        This can be used in the case it contain unrelevant data. Like frame with another shape...
        """
        cache_file = DarkRefsCopy.get_save_file(self._savedir)
        if os.path.exists(cache_file):
            os.remove(cache_file)

    def _clear_cache_data_path(self, file_path: str, data_path: str):
        if not os.path.exists(file_path):
            return
        with h5py.File(file_path, mode="a") as h5f:
            if data_path in h5f:
                del h5f[data_path]

    def set_darks_and_flats_from_scan(self, scan: TomwerScanBase) -> bool:
        has_flats = scan.reduced_flats not in (None, {})
        has_darks = scan.reduced_darks not in (None, {})
        if has_flats and has_darks:
            # if the scan has darks and flats remove directly the cache file
            # else in append mode HDF5 is not removing the dataset and
            # the cache size will continue to increase
            self.clear_cache()
        if not has_flats:
            logger.warning(f"No flat found for {scan}. Unable to copy them")
        else:
            self._clear_cache_data_path(
                file_path=self._flats_url.file_path(),
                data_path=self._flats_url.data_path(),
            )
            dicttoh5(
                scan.reduced_flats,
                h5file=self._flats_url.file_path(),
                h5path=self._flats_url.data_path(),
                mode="a",
                update_mode="replace",
            )
        if not has_darks:
            logger.warning(f"No dark found for {scan}. Unable to copy them")
        else:
            self._clear_cache_data_path(
                file_path=self._darks_url.file_path(),
                data_path=self._darks_url.data_path(),
            )
            dicttoh5(
                scan.reduced_darks,
                h5file=self._darks_url.file_path(),
                h5path=self._darks_url.data_path(),
                mode="a",
                update_mode="replace",
            )
        self._ref_has_been_set(scan=scan)

    def _ref_has_been_set(self, scan):
        pass

    def _copy_to(self, scan):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"{scan} is expected to be an instance of {TomwerScanBase} and not {type(scan)}"
            )

        if self.has_dark_stored():
            copy_h5_dict_darks_to(
                scan=scan,
                darks_url=self._darks_url,
                save=True,
                raise_error_if_url_empty=True,
                overwrite=True,
            )
        if self.has_flat_stored():
            copy_h5_dict_flats_to(
                scan=scan,
                flats_url=self._flats_url,
                save=True,
                raise_error_if_url_empty=True,
                overwrite=True,
            )

    def run(self):
        """
        This is function triggered when a new scan / data is received.
        As explained in issue #184 the behavior is the following:

        * if the scan has already ref files files won't be overwrite
        * if the mode is in `auto` will register last ref file met
        * if the scan has no ref files and refCopy has some register. Will
          create a copy of those, normalized from srCurrent (for flat field)
        """
        scan = data_identifier_to_scan(self.inputs.data)
        if not isinstance(scan, (type(None), TomwerScanBase, dict)):
            raise TypeError(
                f"self.inputs.data is expected to be an instance "
                f"of None, TomwerScanBase or dict. Not {type(scan)}"
            )
        if scan is None or scan.path is None:
            return
        # process dark and ref calculation
        if not self._processOnlyCopy:
            super().run()
        # then copy if necessary
        if not self._processOnlyDkRf:
            if self.contains_dark_or_ref(scan):
                if self.is_on_mode_auto:
                    self.set_darks_and_flats_from_scan(scan)
            if self.has_missing_dark_or_ref(scan):
                self._copy_to(scan)

    def has_flat_or_dark_stored(self) -> bool:
        """

        :return: True if the process has at least registered one flat or one
                 dark
        """
        return self.has_flat_stored() or self.has_dark_stored()

    def has_flat_stored(self) -> bool:
        """

        :return: True if the process has registered at least one ref
        """
        if not os.path.exists(self._save_file):
            return False
        else:
            with open_hdf5(self._save_file) as h5f:
                return self._flats_url.data_path() in h5f

    def has_dark_stored(self) -> bool:
        """

        :return: True if the process has registered at least one dark
        """
        if not os.path.exists(self._save_file):
            return False
        else:
            with open_hdf5(self._save_file) as h5f:
                return self._darks_url.data_path() in h5f

    def contains_dark(self, scan: TomwerScanBase) -> bool:
        """Return True if the scan has already some dark processed"""
        if not isinstance(scan, TomwerScanBase):
            return TypeError(
                f"scan is expected to be an instance of {TomwerScanBase} and not {type(scan)}"
            )
        else:
            return scan.reduced_darks or scan.load_reduced_darks()

    def contains_flat(self, scan: TomwerScanBase):
        """Return True if the scan has already some dark processed"""
        if not isinstance(scan, TomwerScanBase):
            return TypeError(
                f"scan is expected to be an instance of {TomwerScanBase} and not {type(scan)}"
            )
        else:
            return scan.reduced_flats or scan.load_reduced_flats()

    def contains_dark_or_ref(self, scan):
        return self.contains_dark(scan=scan) or self.contains_flat(scan=scan)

    def has_missing_dark_or_ref(self, scan: TomwerScanBase) -> bool:
        """return True if the scan has no ref or no dark registered"""
        assert isinstance(scan, TomwerScanBase)
        return not self.contains_dark(scan) or not self.contains_flat(scan)

    def _signal_done(self, scan):
        assert isinstance(scan, TomwerScanBase)
        raise NotImplementedError("Abstract class")

    def set_mode_auto(self, b):
        self._mode_auto = b

    @property
    def is_on_mode_auto(self):
        return self._mode_auto

    @property
    def refHST_prefix(self):
        return self._refHstPrefix

    @property
    def darkHST_prefix(self):
        return self._darkHstPrefix

    def set_refHST_prefix(self, prefix):
        self._refHstPrefix = prefix

    def set_darkHST_prefix(self, prefix):
        self._darkHstPrefix = prefix



from __future__ import annotations

from silx.gui import qt
from tomwer.core.process.reconstruction.axis.side import Side
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel
from tomwer.gui.utils.qt_utils import block_signals


class _EstimatedCorValidator(qt.QDoubleValidator):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setDecimals(3)

    def validate(self, a0: str, a1: int):
        """validate float or string that could be part of the side values..."""
        for value in Side.values():
            if value.startswith(a0):
                return (qt.QDoubleValidator.Acceptable, a0, a1)
        return super().validate(a0, a1)


class EstimatedCorComboBox(QComboBoxIgnoreWheel):
    """
    Combobox that display the sides available according to the cor algorithm (left, right, center, all)
    and a dedicated item for cor given manually.

    This combobox is also editable and and make sure the 'estimated cor' item is up to date according to the QCombobox current value
    """

    ESTIMATED_COR_ITEM_DATA = "estimated_cor"

    sigEstimatedCorChanged = qt.Signal(object)
    """Emit when the estimated cor changed. Value can be a float (cor value) or a Side"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.addItem("0.0", self.ESTIMATED_COR_ITEM_DATA)
        for side in Side:
            self.addItem(side.value, side)

        self.setValidator(_EstimatedCorValidator())
        self.setEditable(True)

        self.setToolTip(
            """Estimated position of the center of rotation (COR) to be given to the cor algorithms. \n
            If you don't have a fair estimate you can provide only a side
            """
        )
        # connect signal / slot
        self.lineEdit().editingFinished.connect(self._corHasBeenEdited)
        self.currentIndexChanged.connect(self._corChanged)

    def _corHasBeenEdited(self):
        current_cor = self.getCurrentCorValue()
        if isinstance(current_cor, float):
            # keep the item up to date
            self._setCorItemValue(current_cor)
        self.sigEstimatedCorChanged.emit(current_cor)

    def _corChanged(self):
        self.sigEstimatedCorChanged.emit(self.getCurrentCorValue())

    def getCurrentCorValue(self) -> Side | float:
        try:
            return float(self.currentText())
        except ValueError:
            return Side(self.currentText())

    def setCurrentCorValue(self, value: float | Side):
        if isinstance(value, float):
            self._setCorItemValue(value)
        else:
            side = Side(value)
            item = self.findData(side)
            self.setCurrentIndex(item)

    def _setCorItemValue(self, value: float):
        item_index = self.findData(self.ESTIMATED_COR_ITEM_DATA)
        view = self.view()
        hidden = view.isRowHidden(item_index)
        with block_signals(self):
            self.setItemText(item_index, f"{value:.2f}")
            view.setRowHidden(item_index, hidden)

    def get_hidden_sides(self) -> tuple[Side]:
        """Return all sides currently hidden"""
        view = self.view()
        return tuple(
            filter(
                lambda side: view.isRowHidden(self.findData(side)),
                [Side(side) for side in Side],
            )
        )

    def setSidesVisible(self, sides: tuple[Side]):
        """Set side to be visible to the user"""
        sides_visibles = tuple([Side(side) for side in sides])
        view = self.view()
        for side in Side:
            item_idx = self.findData(side)
            view.setRowHidden(item_idx, side not in sides_visibles)

        need_new_current_value = self.getCurrentCorValue() in self.get_hidden_sides()
        if need_new_current_value:
            # if the current side used cannot be used fall back to the estimated cor from motor
            self.setCurrentIndex(self.findData(self.ESTIMATED_COR_ITEM_DATA))

    def setFirstGuessAvailable(self, available: bool):
        """
        For some method (only growing window at the moment) a first guess (estimated cor as a float) cannot be given
        """
        view = self.view()
        item_index = self.findData(self.ESTIMATED_COR_ITEM_DATA)
        view.setRowHidden(item_index, not available)

    def selectFirstGuess(self):
        item_index = self.findData(self.ESTIMATED_COR_ITEM_DATA)
        self.setCurrentIndex(item_index)



# coding: utf-8
from __future__ import annotations


import logging

from silx.gui import qt

from tomwer.core.process.reconstruction.darkref import params as dkrf
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefsTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.utils.sandboxes import RegularExpressionSandBox
from tomwer.synctools.darkref import QDKRFRP

logger = logging.getLogger(__name__)


class DarkRefWidget(qt.QWidget):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    sigScanReady = qt.Signal(TomwerScanBase)
    """signal emitted when a scan is ready"""
    sigProcessingStart = qt.Signal()
    """signal emitted when a processing starts"""
    sigProcessingEnd = qt.Signal()
    """signal emitted when a processing ends"""

    def __init__(self, reconsparams, parent, process_id=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.mainWidget = DarkRefTab(parent=self, reconsparams=reconsparams)
        self.layout().addWidget(self.mainWidget)

        # expose API
        self.setReconsParams = self.mainWidget.setReconsParams

        # set up
        self.mainWidget.setReconsParams(recons_params=reconsparams)

    def _processingStarted(self):
        self.sigProcessingStart.emit()

    def _processingEnded(self):
        self.sigProcessingEnd.emit()

    def setName(self, name):
        self._darkRef.name = name

    def setForceSync(self, sync):
        raise NotImplementedError("")
        self._darkRef.setForceSync(sync)

    def _scanReady(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.sigScanReady.emit(scan)

    @property
    def recons_params(self):
        return self.mainWidget.recons_params


class QDarkRefs(DarkRefsTask):
    def _computationStarts(self):
        pass

    def _computationEnds(self):
        pass


class _WhatCheckBox(qt.QWidget):
    """
    Widget grouping a checkbox and a combobox to know the requested mode
    (None, median, average) for a what (ref, dark)
    """

    sigChanged = qt.Signal(dkrf.ReduceMethod)
    """Signal emitted when the calculation mode change"""

    def __init__(self, parent, text):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self._checkbox = qt.QCheckBox(text=text, parent=self)
        self.layout().addWidget(self._checkbox)
        self._modeCB = qt.QComboBox(parent=self)
        for mode in dkrf.ReduceMethod:
            if mode is dkrf.ReduceMethod.NONE:
                continue
            self._modeCB.addItem(mode.value)
        self._modeCB.setItemData(
            self._modeCB.findText(dkrf.ReduceMethod.MEDIAN.value),
            "Compute the median for each serie",
            qt.Qt.ToolTipRole,
        )
        self._modeCB.setItemData(
            self._modeCB.findText(dkrf.ReduceMethod.MEAN.value),
            "Compute the average for each serie",
            qt.Qt.ToolTipRole,
        )
        self._modeCB.setItemData(
            self._modeCB.findText(dkrf.ReduceMethod.FIRST.value),
            "Pick the first frame for a serie",
            qt.Qt.ToolTipRole,
        )
        self._modeCB.setItemData(
            self._modeCB.findText(dkrf.ReduceMethod.LAST.value),
            "Pick the last frame for a serie",
            qt.Qt.ToolTipRole,
        )

        self.layout().addWidget(self._modeCB)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._checkbox.setChecked(True)
        self._checkbox.toggled.connect(self._modeCB.setEnabled)
        self._checkbox.toggled.connect(self._modeChange)
        self._modeCB.currentIndexChanged.connect(self._modeChange)
        self._updateReconsParam = True
        """Boolean used to know if we have to apply modifications on
        the self._reconsParams (in the case user made modification)
        or not (in the case we are simply reading structure and to avoid
        looping in signals with other QObject)"""

    def getMode(self) -> dkrf.ReduceMethod:
        if self._checkbox.isChecked():
            return dkrf.ReduceMethod.from_value(self._modeCB.currentText())
        else:
            return dkrf.ReduceMethod.NONE

    def getModeName(self):
        return self.getMode().value

    def _modeChange(self, *a, **b):
        self.sigChanged.emit(self.getMode())

    def setMode(self, mode):
        mode = dkrf.ReduceMethod.from_value(mode)
        assert mode in dkrf.ReduceMethod
        self._checkbox.toggled.disconnect(self._modeChange)
        self._modeCB.currentIndexChanged.disconnect(self._modeChange)
        self._checkbox.setChecked(mode is not dkrf.ReduceMethod.NONE)
        if mode is not dkrf.ReduceMethod.NONE:
            index = self._modeCB.findText(mode.value)
            if index < 0:
                logger.error(f"index for {mode.value} is not recognized")
            else:
                self._modeCB.setCurrentIndex(index)
        self._checkbox.toggled.connect(self._modeChange)
        self._modeCB.currentIndexChanged.connect(self._modeChange)
        self.sigChanged.emit(self.getMode())


class _TabGeneral(qt.QWidget):
    """Widget with the general information for dark and ref process"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self._grpWhat = qt.QGroupBox("what", parent=self)
        self._grpWhat.setLayout(qt.QVBoxLayout())
        self._darkWCB = _WhatCheckBox(parent=self._grpWhat, text="dark")
        self._refWCB = _WhatCheckBox(parent=self._grpWhat, text="flat")
        self._grpWhat.layout().addWidget(self._darkWCB)
        self.sigDarkChanged = self._darkWCB.sigChanged
        self._grpWhat.layout().addWidget(self._refWCB)
        self.sigRefChanged = self._refWCB.sigChanged
        self.layout().addWidget(self._grpWhat)

        self._grpOptions = qt.QGroupBox("options", parent=self)
        self._grpOptions.setLayout(qt.QVBoxLayout())
        self._rmOptionCB = qt.QCheckBox(
            parent=self._grpOptions,
            text="remove raw EDF files when done (for spec only)",
        )
        self.sigRmToggled = self._rmOptionCB.toggled
        self._skipOptionCB = qt.QCheckBox(
            parent=self._grpOptions, text="skip if already existing"
        )
        self.sigSkipToggled = self._skipOptionCB.toggled
        self._grpOptions.layout().addWidget(self._rmOptionCB)
        self._grpOptions.layout().addWidget(self._skipOptionCB)
        self.layout().addWidget(self._grpOptions)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)

        self.layout().addWidget(spacer)


class _TabExpert(qt.QWidget):
    """Expert process for dark and ref"""

    sigDarkPatternEdited = qt.Signal(str)
    sigRefPatternEdited = qt.Signal(str)

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self._patternsWidget = qt.QWidget(parent=self)
        self._patternsWidget.setLayout(qt.QGridLayout())

        self._patternsWidget.layout().addWidget(
            qt.QLabel("pyhst dark file pattern", parent=self._patternsWidget), 0, 0
        )
        self._darkLE = qt.QLineEdit(parent=self._patternsWidget)
        self._darkLE.setToolTip(DarkRefsTask.getDarkPatternTooltip())
        self._darkLE.editingFinished.connect(self._darkPatternEdited)
        self._patternsWidget.layout().addWidget(self._darkLE, 0, 1)
        self._patternsWidget.layout().addWidget(
            qt.QLabel("pyhst flat file pattern", parent=self._patternsWidget), 1, 0
        )
        self._refLE = qt.QLineEdit(parent=self._patternsWidget)
        self._refLE.setToolTip(DarkRefsTask.getRefPatternTooltip())
        self._refLE.editingFinished.connect(self._refPatternEdited)

        self._patternsWidget.layout().addWidget(self._refLE, 1, 1)

        self.layout().addWidget(self._patternsWidget)

        textExtraInfo = (
            "note: to have more information about pattern usage \n"
            "see tooltips over dark and flat field patterns."
            "\nYou can also see help to have advance"
            "information"
        )
        labelNote = qt.QLabel(parent=self, text=textExtraInfo)
        labelNote.setSizePolicy(qt.QSizePolicy.Preferred, qt.QSizePolicy.Minimum)
        self.layout().addWidget(labelNote)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)

        self.layout().addWidget(spacer)

    def _darkPatternEdited(self):
        self.sigDarkPatternEdited.emit(self._darkLE.text())

    def _refPatternEdited(self):
        self.sigRefPatternEdited.emit(self._refLE.text())


class DarkRefTab(qt.QTabWidget):
    # TODO: should modify the reconsparams to

    def __init__(self, parent, reconsparams):
        self.recons_params = None
        qt.QTabWidget.__init__(self, parent)
        self.tabGeneral = _TabGeneral(parent=self)
        self.addTab(self.tabGeneral, "general")
        self.tabExpert = _TabExpert(parent=self)
        self.addTab(self.tabExpert, "expert")
        self.tabSandBox = RegularExpressionSandBox(parent=self, pattern="")
        extraSandBoxInfo = qt.QLabel(
            "This sand box allow you to play with"
            "regular expressions and check the "
            "validity of a defined pattern versus "
            "acquisition names. This way you can be "
            "insure of the regular expression "
            "behavior",
            parent=self.tabSandBox,
        )
        extraSandBoxInfo.setWordWrap(True)
        self.tabSandBox.layout().insertWidget(0, extraSandBoxInfo)
        self.tabSandBox.setToolTip("play with regular expression")
        self.addTab(self.tabSandBox, "re sandbox")

        self._makeConnection()
        self.setReconsParams(recons_params=reconsparams)

    def _disconnectToReconsParams(self):
        assert self.recons_params
        try:
            self.tabGeneral.sigDarkChanged.disconnect(
                self.recons_params.__class__.dark_calc_method.setter
            )
            self.tabGeneral.sigRefChanged.disconnect(
                self.recons_params.__class__.flat_calc_method.setter
            )
            self.tabGeneral.sigRmToggled.disconnect(self.recons_params._set_remove_opt)
            self.tabGeneral.sigSkipToggled.disconnect(
                self.recons_params._set_skip_if_exist
            )

            self.tabExpert.sigDarkPatternEdited.disconnect(
                self.recons_params.__class__.dark_pattern.setter
            )
            self.tabExpert.sigRefPatternEdited.disconnect(
                self.recons_params.__class__.ref_pattern.setter
            )
            self.recons_params.sigChanged.disconnect(self._updateReconsParams)
        except Exception:
            pass

    def _connectToReconsParams(self):
        assert self.recons_params
        self.tabGeneral.sigDarkChanged.connect(
            self.recons_params.__class__.dark_calc_method.setter
        )
        self.tabGeneral.sigRefChanged.connect(
            self.recons_params.__class__.flat_calc_method.setter
        )
        self.tabGeneral.sigRmToggled.connect(self.recons_params._set_remove_opt)
        self.tabGeneral.sigSkipToggled.connect(self.recons_params._set_skip_if_exist)

        self.tabExpert.sigDarkPatternEdited.connect(
            self.recons_params.__class__.dark_pattern.setter
        )
        self.tabExpert.sigRefPatternEdited.connect(
            self.recons_params.__class__.flat_pattern.setter
        )
        self.recons_params.sigChanged.connect(self._updateReconsParams)

    def _updateReconsParams(self):
        self.loadStructs(self.recons_params.to_dict())

    def setReconsParams(self, recons_params):
        if not isinstance(recons_params, QDKRFRP):
            raise TypeError
        _recons_params = recons_params

        if self.recons_params:
            self._disconnectToReconsParams()

        self.recons_params = _recons_params
        self._updateReconsParams()
        self._connectToReconsParams()

    def _makeConnection(self):
        self.tabGeneral._refWCB.sigChanged.connect(self._refCalcModeChanged)
        self.tabGeneral._darkWCB.sigChanged.connect(self._darkCalcModeChanged)
        self.tabGeneral._rmOptionCB.toggled.connect(self._rmOptChanged)
        self.tabGeneral._skipOptionCB.toggled.connect(self._skipOptChanged)
        self.tabExpert._darkLE.editingFinished.connect(self._darkPatternChanged)
        self.tabExpert._refLE.editingFinished.connect(self._refPatternChanged)

    def loadStructs(self, structs):
        def warningKeyNotHere(key):
            logger.warning(
                "%s key not present in the given struct, "
                "cannot load value for it." % key
            )

        assert isinstance(structs, dict)

        if "DARKCAL" not in structs:
            warningKeyNotHere("DARKCAL")
        else:
            self.setDarkMode(structs["DARKCAL"])

        if "REFSCAL" not in structs:
            warningKeyNotHere("REFSCAL")
        else:
            self.setRefMode(structs["REFSCAL"])

        if "REFSOVE" not in structs:
            warningKeyNotHere("REFSOVE")
        else:
            self.setSkipOption(not structs["REFSOVE"])

        if "REFSRMV" not in structs:
            warningKeyNotHere("REFSRMV")
        else:
            self.setRemoveOption(structs["REFSRMV"])

        if "RFFILE" not in structs:
            warningKeyNotHere("RFFILE")
        else:
            self.setRefPattern(structs["RFFILE"])

        if "DKFILE" not in structs:
            warningKeyNotHere("DKFILE")
        else:
            self.setDarkPattern(structs["DKFILE"])

    def setRemoveOption(self, rm):
        self.tabGeneral._rmOptionCB.setChecked(rm)

    def setSkipOption(self, skip):
        self.tabGeneral._skipOptionCB.setChecked(skip)

    def setDarkMode(self, mode):
        self.tabGeneral._darkWCB.setMode(mode)

    def setRefMode(self, mode):
        self.tabGeneral._refWCB.setMode(mode)

    def setRefPattern(self, pattern):
        self.tabExpert._refLE.setText(pattern)

    def setDarkPattern(self, pattern):
        self.tabExpert._darkLE.setText(pattern)

    def _rmOptChanged(self):
        value = self.tabGeneral._rmOptionCB.isChecked()
        self.recons_params._set_remove_opt(value)

    def _skipOptChanged(self):
        value = self.tabGeneral._skipOptionCB.isChecked()
        self.recons_params._set_skip_if_exist(value)

    def _refPatternChanged(self):
        value = self.tabExpert._refLE.text()
        self.recons_params["RFFILE"] = value

    def _darkPatternChanged(self):
        value = self.tabExpert._darkLE.text()
        self.recons_params["DKFILE"] = value

    def _darkCalcModeChanged(self):
        value = self.tabGeneral._darkWCB.getMode()
        self.recons_params["DARKCAL"] = value

    def _refCalcModeChanged(self):
        value = self.tabGeneral._refWCB.getMode()
        self.recons_params["REFSCAL"] = value



# coding: utf-8
"""
contains gui relative to intensity normalization
"""

from __future__ import annotations


import weakref

from silx.gui import qt
from silx.gui.dialog.DataFileDialog import DataFileDialog
from silx.gui.plot.items.roi import HorizontalRangeROI, RectangleROI
from silx.gui.plot.tools.roi import RegionOfInterestManager
from silx.io.url import DataUrl
from tomoscan.normalization import Method

from tomwer.core.process.reconstruction.normalization import params as _normParams
from tomwer.core.process.reconstruction.normalization.params import _ValueSource
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.scores.control import ControlWidget
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.visualization.dataviewer import DataViewer
from tomwer.gui.visualization.sinogramviewer import SinogramViewer as _SinogramViewer


class SinoNormWindow(qt.QMainWindow):
    sigConfigurationChanged = qt.Signal()
    """signal emit when the configuration change"""

    def __init__(self, parent, backend=None):
        qt.QMainWindow.__init__(self, parent)
        self._scan = None
        self.setWindowFlags(qt.Qt.Widget)

        # central widget
        self._centralWidget = _Viewer(self, backend=backend)
        self.setCentralWidget(self._centralWidget)

        # control widget (options + ctrl buttons)
        self._dockWidgetWidget = qt.QWidget(self)
        self._dockWidgetWidget.setLayout(qt.QVBoxLayout())
        self._optsWidget = _NormIntensityOptions(self)
        self._dockWidgetWidget.layout().addWidget(self._optsWidget)
        self._spacer = qt.QWidget(self)
        self._spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._dockWidgetWidget.layout().addWidget(self._spacer)
        self._crtWidget = _NormIntensityControl(self)
        self._dockWidgetWidget.layout().addWidget(self._crtWidget)

        # dock widget
        self._dockWidgetCtrl = qt.QDockWidget(parent=self)
        self._dockWidgetCtrl.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._dockWidgetCtrl.setWidget(self._dockWidgetWidget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidgetCtrl)

        # connect signal / slot
        # self._optsWidget.sigModeChanged.connect(self._modeChanged)
        self._optsWidget.sigValueUpdated.connect(self.setResult)
        self._optsWidget.sigConfigurationChanged.connect(self.sigConfigurationChanged)
        self._optsWidget.sigSourceChanged.connect(self._sourceChanged)
        self._crtWidget.sigValidateRequest.connect(self._validated)

        # set up
        self._centralWidget._updateSinogramROI()
        self._modeChanged()

    def _hideLockButton(self):
        self._optsWidget._hideLockButton()

    def _validated(self):
        pass

    def getConfiguration(self) -> dict:
        return self._optsWidget.getConfiguration()

    def setConfiguration(self, config: dict):
        self._optsWidget.setConfiguration(config=config)

    def setCurrentMethod(self, method):
        self._optsWidget.setCurrentMethod(method=method)

    def getCurrentMethod(self):
        return self._optsWidget.getCurrentMethod()

    def setCurrentSource(self, source):
        self._optsWidget.setCurrentSource(source=source)

    def getCurrentSource(self):
        return self._optsWidget.getCurrentSource()

    def _modeChanged(self):
        self._sourceChanged()

    def _sourceChanged(self):
        source = self.getCurrentSource()
        method = self.getCurrentMethod()
        scan = self.getScan()

        methods_using_manual_roi = (Method.DIVISION, Method.SUBTRACTION)

        self._centralWidget.setManualROIVisible(
            source is _ValueSource.MANUAL_ROI and method in methods_using_manual_roi
        )
        if scan:
            methods_requesting_calculation = (Method.DIVISION, Method.SUBTRACTION)
            if method in methods_requesting_calculation:
                # if the normed sinogram can be obtained `directly`
                if source in (_ValueSource.MANUAL_SCALAR, _ValueSource.DATASET):
                    scan.intensity_normalization = self.getCurrentMethod().value
                    extra_info = self.getExtraArgs()
                    extra_info.update(
                        {
                            "tomwer_processing_res_code": True,
                            "source": source.value,
                        }
                    )
                    scan.intensity_normalization.set_extra_infos(extra_info)
        self._centralWidget._updateSinogramROI()

    def getScan(self):
        if self._scan is not None:
            return self._scan()
        else:
            return None

    def setScan(self, scan: TomwerScanBase | None):
        self._scan = weakref.ref(scan)
        self._centralWidget.setScan(scan=scan)
        self._optsWidget.setScan(scan=scan)

    def getExtraArgs(self) -> dict:
        return self._optsWidget.getExtraInfos()

    def getROI(self):
        return self._centralWidget.getROI()

    def setROI(self, start_x, end_x, start_y, end_y):
        self._centralWidget.setROI(
            start_x=start_x, end_x=end_x, start_y=start_y, end_y=end_y
        )

    def setResult(self, result):
        self._crtWidget.setResult(result)

    def clear(self):
        self._crtWidget.clear()

    def isLocked(self):
        return self._optsWidget.isLocked()

    def setLocked(self, locked):
        self._optsWidget.setLocked(locked)


class _Viewer(qt.QTabWidget):
    def __init__(self, parent, backend=None):
        if not isinstance(parent, SinoNormWindow):
            raise TypeError("Expect a NormIntensityWindow as parrent")
        qt.QTabWidget.__init__(self, parent)
        self._projView = _ProjPlotWithROI(parent=self, backend=backend)
        self.addTab(self._projView, "projection view")
        self._sinoView = SinogramViewer(parent=self, backend=backend)
        self.addTab(self._sinoView, "sinogram view")

        # connect signal / Slot
        self._sinoView.sigSinogramLineChanged.connect(self._projView.setSinogramLine)
        self._sinoView.sigSinoLoadEnded.connect(self._updateSinogramROI)
        self._projView.sigROIChanged.connect(self._updateSinogramROI)

    def setScan(self, scan: TomwerScanBase | None):
        """

        :param scan: scan to handle
        :return:
        """
        self._projView.setScan(scan)
        self._sinoView.setScan(scan, update=False)

    def _updateSinogramROI(self):
        source = self.parent().getCurrentSource()
        method = self.parent().getCurrentMethod()

        display_sino_roi = source is _ValueSource.MANUAL_ROI and method in (
            Method.DIVISION,
            Method.SUBTRACTION,
        )

        if display_sino_roi:
            roi = self._projView.getROI()
            sinogram_line = self._sinoView.getLine()
            y_min = roi.getOrigin()[1]
            y_max = roi.getOrigin()[1] + roi.getSize()[1]
            if y_min <= sinogram_line <= y_max:
                x_min = roi.getOrigin()[0]
                x_max = roi.getOrigin()[0] + roi.getSize()[0]
                self._sinoView.setROIRange(x_min, x_max)
                self._sinoView.setROIVisible(True)
            else:
                self._sinoView.setROIVisible(False)
        else:
            self._sinoView.setROIVisible(False)

    def setManualROIVisible(self, visible):
        self._projView.setManualROIVisible(visible=visible)
        self._sinoView.setROIVisible(visible=visible)

    def getROI(self):
        return self._projView.getROI()

    def setROI(self, start_x, end_x, start_y, end_y):
        self._projView.setROI(
            start_x=start_x, end_x=end_x, start_y=start_y, end_y=end_y
        )


class _ProjPlotWithROI(DataViewer):
    """DataViewer specialized on projections. Embed a RectangleROI"""

    sigROIChanged = qt.Signal()
    """signal emit when ROI change"""

    def __init__(self, *args, **kwargs):
        DataViewer.__init__(self, *args, **kwargs, show_overview=False)
        self._sinogramLine = 0
        self._roiVisible = False
        self.setScanInfoVisible(False)
        self.setDisplayMode("projections-radios")
        self.setDisplayModeVisible(False)

        dw = self.getUrlListDockWidget()
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, dw)
        dw.setFeatures(qt.QDockWidget.DockWidgetMovable)

        # add ROI
        self._roiManager = RegionOfInterestManager(self.getPlotWidget())
        self._roi = RectangleROI()
        self._roi.setName("ROI")
        self._roi.setLineStyle("-.")
        self._roi.setGeometry(origin=(0, 0), size=(200, 200))
        self._roi.setEditable(True)
        self._roi.setVisible(True)
        self._roiManager.addRoi(self._roi)

        # connect signal / slot
        self.getPlotWidget().sigActiveImageChanged.connect(self._updateSinogramLine)
        self.getPlotWidget().sigActiveImageChanged.connect(self._updateROI)
        self._roi.sigEditingFinished.connect(self._roiChanged)

    def getROI(self):
        return self._roi

    def setROI(self, start_x, end_x, start_y, end_y):
        self._roi.setOrigin((start_x, start_y))
        self._roi.setSize((end_x - start_x, end_y - start_y))

    def setManualROIVisible(self, visible):
        self._roiVisible = visible
        self._updateROIVisibility()

    def _updateROIVisibility(self):
        self._roi.setVisible(self._roiVisible)

    def _roiChanged(self):
        self.sigROIChanged.emit()

    def setSinogramLine(self, line):
        self._sinogramLine = line
        self._updateSinogramLine()

    def _updateROI(self):
        """ImageStack clean the plot which bring item removal. This is
        why we need to add them back"""
        if self._roi is not None:
            for item in self._roi.getItems():
                if item not in self.getPlotWidget().getItems():
                    self.getPlotWidget().addItem(item)

    def _updateSinogramLine(self):
        self._roiManager = RegionOfInterestManager(self.getPlotWidget())

        self.getPlotWidget().addYMarker(
            y=self._sinogramLine,
            legend="sinogram_line",
            text="sinogram line",
            color="blue",
            selectable=False,
        )
        sino_marker = self.getPlotWidget()._getMarker("sinogram_line")
        if sino_marker:
            sino_marker.setLineStyle("--")

    def clear(self):
        super().clear()
        self.getPlotWidget().removeMarker("sinogram_line")


class SinogramViewer(_SinogramViewer):
    """ "Sinogram viewer but adapated for Intensity normalization"""

    sigSinogramLineChanged = qt.Signal(int)
    """signal emit when the selected sinogram line changes"""

    def __init__(self, *args, **kwargs):
        _SinogramViewer.__init__(self, *args, **kwargs)

        dockWidget = self.getOptionsDockWidget()
        self.addDockWidget(qt.Qt.TopDockWidgetArea, dockWidget)
        dockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)

        # change ApplyButton name and icon
        self._loadButton = self._options._buttons.button(qt.QDialogButtonBox.Apply)
        self._loadButton.setText("load")
        style = qt.QApplication.style()
        self._loadButton.setIcon(style.standardIcon(qt.QStyle.SP_BrowserReload))

        # ROI
        self._roiManager = RegionOfInterestManager(self.getPlotWidget())
        self._roi = HorizontalRangeROI()
        self._roi.setRange(0, 0)
        self._roi.setVisible(True)
        self._roi.setColor((250, 50, 50, 150))
        self._roi.setLineWidth(1.5)
        self._roi.setLineStyle("-.")
        self._roiManager.addRoi(self._roi)

        # connect signal / Slots
        self._options._lineSB.valueChanged.connect(self._sinogramLineChanged)
        self.sigSinogramLineChanged.connect(self._plot.clear)
        self.getPlotWidget().sigActiveImageChanged.connect(self._updateROI)

    def _sinogramLineChanged(self):
        self.sigSinogramLineChanged.emit(self.getLine())

    def setROIRange(self, x_min, x_max):
        self._roi.setRange(x_min, x_max)

    def setROIVisible(self, visible):
        self._roi.setVisible(visible)

    def _updateROI(self):
        """ImageStack clean the plot which bring item removal. This is
        why we need to add them back"""
        if self._roi is not None:
            for item in self._roi.getItems():
                if item not in self.getPlotWidget().getItems():
                    self.getPlotWidget().addItem(item)

    def getPlotWidget(self):
        return self._plot

    def _updatePlot(self, sinogram):
        self.getPlotWidget().addImage(data=sinogram)
        self.getPlotWidget().replot()


class _NormIntensityOptions(qt.QWidget):
    sigValueCanBeLocked = qt.Signal(bool)

    sigProcessingRequested = qt.Signal()
    """Signal emit when the processing is requested"""

    sigModeChanged = qt.Signal()
    """signal emitted when the mode change"""

    sigSourceChanged = qt.Signal()
    """signal emitted when the source change"""

    sigValueUpdated = qt.Signal(object)
    """Signal emit when user defines manually the value"""

    sigConfigurationChanged = qt.Signal()
    """Signal emit when the configuration changes"""

    def __init__(self, parent):
        if not isinstance(parent, SinoNormWindow):
            raise TypeError(
                "parent is expected to be an instance of " "NormIntensityWindow "
            )
        qt.QWidget.__init__(self, parent)
        self._getROI = self.parent().getROI
        self.setLayout(qt.QGridLayout())
        # mode
        self._modeCB = qt.QComboBox(self)
        for mode in Method:
            if mode in (Method.LSQR_SPLINE,):
                continue
            else:
                self._modeCB.addItem(mode.value)
        self.layout().addWidget(qt.QLabel("mode:", self), 0, 0, 1, 1)
        self.layout().addWidget(self._modeCB, 0, 1, 1, 1)
        self._lockButton = PadlockButton(self)
        self._lockButton.setFixedWidth(25)
        self.layout().addWidget(self._lockButton, 0, 2, 1, 1)
        # source
        self._sourceCB = qt.QComboBox(self)
        for mode in _ValueSource:
            if mode == _ValueSource.NONE:
                # filter this value because does not have much sense for the GUI
                continue
            if mode in (_ValueSource.AUTO_ROI, _ValueSource.MONITOR):
                continue
            self._sourceCB.addItem(mode.value)
        self._sourceLabel = qt.QLabel("source:", self)
        self.layout().addWidget(self._sourceLabel, 1, 0, 1, 1)
        self.layout().addWidget(self._sourceCB, 1, 1, 1, 1)
        # method
        self._optsMethod = qt.QGroupBox(self)
        self._optsMethod.setTitle("options")
        self._optsMethod.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self._optsMethod, 2, 0, 1, 3)
        # intensity calculation options
        self._intensityCalcOpts = _NormIntensityCalcOpts(self)
        self._optsMethod.layout().addWidget(self._intensityCalcOpts)
        # dataset widget
        self._datasetWidget = _NormIntensityDatasetWidget(self)
        self._optsMethod.layout().addWidget(self._datasetWidget)
        # scalar value
        self._scalarValueWidget = _NormIntensityScalarValue(self)
        self.layout().addWidget(self._scalarValueWidget, 3, 0, 1, 3)
        # buttons
        self._buttonsGrp = qt.QWidget(self)
        self._buttonsGrp.setLayout(qt.QGridLayout())
        self._buttonsGrp.layout().setContentsMargins(0, 0, 0, 0)
        self._computeButton = qt.QPushButton("compute", self)
        self._buttonsGrp.layout().addWidget(self._computeButton, 0, 1, 1, 1)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._buttonsGrp.layout().addWidget(spacer)
        self.layout().addWidget(self._buttonsGrp, 4, 0, 1, 3)

        self._modeChanged()

        # connect signal / slot
        self._modeCB.currentIndexChanged.connect(self._modeChanged)
        self._modeCB.currentIndexChanged.connect(self.sigConfigurationChanged)
        self._sourceCB.currentIndexChanged.connect(self._sourceChanged)
        self._sourceCB.currentIndexChanged.connect(self.sigConfigurationChanged)
        self._computeButton.released.connect(self.sigProcessingRequested)
        self._scalarValueWidget.sigValueChanged.connect(self._valueUpdated)
        self._scalarValueWidget.sigValueChanged.connect(self.sigConfigurationChanged)
        self._datasetWidget.sigConfigurationChanged.connect(
            self.sigConfigurationChanged
        )
        self._intensityCalcOpts.sigConfigurationChanged.connect(
            self.sigConfigurationChanged
        )
        self._lockButton.toggled.connect(self._lockChanged)
        self._lockButton.toggled.connect(self.sigConfigurationChanged)

    def _sourceChanged(self):
        source = self.getCurrentSource()
        method = self.getCurrentMethod()
        interactive_methods = (Method.DIVISION, Method.SUBTRACTION)
        interactive_sources = (
            _ValueSource.MANUAL_ROI,
            _ValueSource.AUTO_ROI,
            _ValueSource.DATASET,
        )
        self._datasetWidget.setVisible(source == _ValueSource.DATASET)
        self.setManualROIVisible(source == _ValueSource.MANUAL_ROI)
        self._optsMethod.setVisible(
            method in interactive_methods and source in interactive_sources
        )
        self._intensityCalcOpts.setCalculationFctVisible(
            method in interactive_methods
            and source in interactive_sources
            and source is not _ValueSource.DATASET
        )
        self._scalarValueWidget.setVisible(source == _ValueSource.MANUAL_SCALAR)
        self._buttonsGrp.setVisible(
            method in interactive_methods and source in interactive_sources
        )

        self.sigSourceChanged.emit()

    def _lockChanged(self):
        self._scalarValueWidget.setEnabled(not self.isLocked())
        self._datasetWidget.setEnabled(not self.isLocked())
        self._intensityCalcOpts.setEnabled(not self.isLocked())
        self._modeCB.setEnabled(not self.isLocked())
        self._computeButton.setEnabled(not self.isLocked())

    def isLocked(self):
        return self._lockButton.isLocked()

    def setLocked(self, locked):
        self._lockButton.setChecked(locked)

    def _hideLockButton(self):
        self._lockButton.hide()

    def getCurrentMethod(self):
        return Method.from_value(self._modeCB.currentText())

    def setCurrentMethod(self, method):
        method = Method.from_value(method)
        idx = self._modeCB.findText(method.value)
        self._modeCB.setCurrentIndex(idx)

    def getCurrentSource(self):
        if self.getCurrentMethod() in (Method.DIVISION, Method.SUBTRACTION):
            return _ValueSource.from_value(self._sourceCB.currentText())
        else:
            return _ValueSource.NONE

    def setCurrentSource(self, source):
        source = _ValueSource.from_value(source)
        idx = self._sourceCB.findText(source.value)
        self._sourceCB.setCurrentIndex(idx)

    def _modeChanged(self, *args, **kwargs):
        mode = self.getCurrentMethod()
        mode_with_calculations = (Method.DIVISION, Method.SUBTRACTION)
        self._intensityCalcOpts.setVisible(mode in mode_with_calculations)
        self._sourceCB.setVisible(mode in mode_with_calculations)
        self._sourceLabel.setVisible(mode in mode_with_calculations)
        self._sourceChanged()
        self.sigModeChanged.emit()
        self.sigSourceChanged.emit()

    def _valueUpdated(self, *args):
        self.sigValueUpdated.emit(args)

    def setManualROIVisible(self, visible):
        pass

    def getConfiguration(self) -> dict:
        return _normParams.SinoNormalizationParams(
            method=self.getCurrentMethod(),
            source=self.getCurrentSource(),
            extra_infos=self.getExtraInfos(),
        ).to_dict()

    def setConfiguration(self, config: dict):
        params = _normParams.SinoNormalizationParams.from_dict(config)
        self.setCurrentMethod(params.method)
        extra_infos = params.extra_infos
        if (
            "start_x" in extra_infos
            and "start_y" in extra_infos
            and "end_x" in extra_infos
            and "end_y" in extra_infos
        ):
            start_x = extra_infos["start_x"]
            start_y = extra_infos["start_y"]
            end_x = extra_infos["end_x"]
            end_y = extra_infos["end_y"]
            self._getROI().setOrigin((start_x, start_y))
            self._getROI().setSize((end_x - start_x, end_y - start_y))
        if "calc_fct" in extra_infos:
            self._intensityCalcOpts.setCalculationFct(extra_infos["calc_fct"])
        if "calc_area" in extra_infos:
            self._intensityCalcOpts.setCalculationArea(extra_infos["calc_area"])
        if "calc_method" in extra_infos:
            self._intensityCalcOpts.setCalculationMethod(extra_infos["calc_method"])
        if params.source is _ValueSource.MANUAL_SCALAR:
            if "value" in extra_infos:
                self._scalarValueWidget.setValue(extra_infos["value"])

    def setScan(self, scan):
        self._datasetWidget.setScan(scan=scan)

    def getExtraInfos(self):
        method = self.getCurrentMethod()
        source = self.getCurrentSource()
        if method in (Method.CHEBYSHEV, Method.NONE):
            return {}
        else:
            if source is _ValueSource.MANUAL_SCALAR:
                return {"value": self._scalarValueWidget.getValue()}
            elif source is _ValueSource.AUTO_ROI:
                raise NotImplementedError("auto roi not implemented yet")
            elif source is _ValueSource.MANUAL_ROI:
                roi = self._getROI()
                return {
                    "start_x": roi.getOrigin()[0],
                    "end_x": roi.getOrigin()[0] + roi.getSize()[0],
                    "start_y": roi.getOrigin()[1],
                    "end_y": roi.getOrigin()[1] + roi.getSize()[1],
                    "calc_fct": self._intensityCalcOpts.getCalculationFct().value,
                }
            elif source is _ValueSource.DATASET:
                return {
                    "dataset_url": self._datasetWidget.getDatasetUrl().path(),
                }
            else:
                raise ValueError(f"unhandled source: {source} for method {method}")


class _NormIntensityCalcOpts(qt.QWidget):
    """Options to compute the norm intensity"""

    sigConfigurationChanged = qt.Signal()
    """Signal emitted when configuration changes"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())
        # calculation function
        self._calculationModeCB = qt.QComboBox(self)
        for fct in _normParams._ValueCalculationFct.values():
            self._calculationModeCB.addItem(fct)
        self._calculationModeLabel = qt.QLabel("calculation fct", self)
        self.layout().addRow(self._calculationModeLabel, self._calculationModeCB)

        # connect signal / slot
        self._calculationModeCB.currentIndexChanged.connect(
            self.sigConfigurationChanged
        )

    def setCalculationFctVisible(self, visible):
        self._calculationModeLabel.setVisible(visible)
        self._calculationModeCB.setVisible(visible)

    def getCalculationFct(self):
        return _normParams._ValueCalculationFct.from_value(
            self._calculationModeCB.currentText()
        )

    def setCalculationFct(self, fct):
        idx = self._calculationModeCB.findText(
            _normParams._ValueCalculationFct.from_value(fct)
        )
        self._calculationModeCB.setCurrentIndex(idx)


class _NormIntensityControl(ControlWidget):
    def __init__(self, parent=None):
        ControlWidget.__init__(self, parent)
        self._resultWidget = qt.QWidget(self)
        self._resultWidget.setLayout(qt.QFormLayout())
        self._result = None

        self._resultQLE = qt.QLineEdit("", self)
        self._resultWidget.layout().addRow("value:", self._resultQLE)
        self._resultQLE.setReadOnly(True)
        self.layout().insertWidget(0, self._resultWidget)

        self._computeBut.hide()

    def setResult(self, result):
        self._result = result
        if isinstance(result, tuple):
            result = ",".join([str(element) for element in result])
        self._resultQLE.setText(str(result))

    def getResult(self):
        return self._result

    def clear(self):
        self._resultQLE.clear()


class _NormIntensityScalarValue(qt.QWidget):
    sigValueChanged = qt.Signal(float)
    """emit when the scalar value to norm intensity change"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QFormLayout())
        self._QLE = qt.QLineEdit("0.0", self)
        validator = qt.QDoubleValidator(parent=self)
        self._QLE.setValidator(validator)
        self.layout().addRow("value", self._QLE)

        # connect signal / slot
        self._QLE.editingFinished.connect(self._valueChanged)

    def setValue(self, value):
        self._QLE.setText(str(value))

    def getValue(self):
        return float(self._QLE.text())

    def _valueChanged(self):
        self.sigValueChanged.emit(self.getValue())


class _NormIntensityDatasetWidget(qt.QWidget):
    _FILE_PATH_LOCAL_VALUE = "scan master file"

    sigConfigurationChanged = qt.Signal()
    """emit when configuration change"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self._lastGlobalPath = None
        self._scan = None

        self.setLayout(qt.QGridLayout())
        # file scope
        self._fileScopeGB = qt.QGroupBox("file scope", self)
        self._fileScopeGB.setLayout(qt.QVBoxLayout())
        self._buttonGrpBox = qt.QButtonGroup(self)
        self._globalRB = qt.QRadioButton(_normParams._DatasetScope.GLOBAL.value, self)
        self._buttonGrpBox.addButton(self._globalRB)
        self._globalRB.setToolTip("Global dataset. Will be constant with time")
        self._fileScopeGB.layout().addWidget(self._globalRB)
        self._localRB = qt.QRadioButton(_normParams._DatasetScope.LOCAL.value, self)
        self._buttonGrpBox.addButton(self._localRB)
        self._localRB.setToolTip(
            "Local dataset."
            "Must be contained in the NXtomo entry "
            "provided (not compatible with EDF)."
        )
        self._fileScopeGB.layout().addWidget(self._localRB)
        self.layout().addWidget(self._fileScopeGB, 0, 0, 3, 3)
        # file_path
        self._filePathLabel = qt.QLabel("file path", self)
        self.layout().addWidget(self._filePathLabel, 3, 0, 1, 1)
        self._filePathQLE = qt.QLineEdit("", self)
        self.layout().addWidget(self._filePathQLE, 3, 1, 1, 1)
        self._selectFileButton = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectFileButton, 3, 2, 1, 1)
        # data path
        self._dataPathLabel = qt.QLabel("data path", self)
        self.layout().addWidget(self._dataPathLabel, 4, 0, 1, 1)
        self._dataPathQLE = qt.QLineEdit("", self)
        self.layout().addWidget(self._dataPathQLE, 4, 1, 1, 1)
        self._selectDataPathButton = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectDataPathButton, 4, 2, 1, 1)

        # set up
        self._localRB.setChecked(True)
        self._updateFilePathVisibility()

        # connect signal / slot
        self._buttonGrpBox.buttonToggled.connect(self._updateFilePathVisibility)
        self._selectFileButton.released.connect(self._selectFile)
        self._selectDataPathButton.released.connect(self._selectDataPath)
        self._buttonGrpBox.buttonReleased.connect(self._configurationChanged)
        self._filePathQLE.editingFinished.connect(self._configurationChanged)
        self._dataPathQLE.editingFinished.connect(self._configurationChanged)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigurationChanged.emit()

    def setScan(self, scan):
        if scan is not None:
            self._scan = weakref.ref(scan)
        return None

    def getScan(self):
        if self._scan is not None:
            return self._scan()
        else:
            return None

    def getMode(self) -> _normParams._DatasetScope:
        if self._localRB.isChecked():
            return _normParams._DatasetScope.LOCAL
        else:
            return _normParams._DatasetScope.GLOBAL

    def getDataPath(self):
        return self._dataPathQLE.text()

    def setDataPath(self, path):
        self._dataPathQLE.setText(path)

    def getGlobalFilePath(self) -> str:
        return self._filePathQLE.text()

    def setGlobalFilePath(self, file_path):
        self._filePathQLE.setText(file_path)

    def getDatasetUrl(self) -> DataUrl:
        if self.getMode() is _normParams._DatasetScope.LOCAL:
            if self.getScan() is not None:
                scan = self.getScan()
                file_path = scan.master_file
            else:
                file_path = None
        else:
            file_path = self.getGlobalFilePath()
        data_path = self.getDataPath()
        if file_path is not None and file_path.lower().endswith("edf"):
            scheme = "fabio"
        else:
            scheme = "silx"

        return DataUrl(
            file_path=file_path,
            data_path=data_path,
            scheme=scheme,
        )

    def setDatasetUrl(self, url: DataUrl):
        raise NotImplementedError("")

    def _updateFilePathVisibility(self):
        self._filePathQLE.setReadOnly(self.getMode() == _normParams._DatasetScope.LOCAL)
        self._filePathQLE.setEnabled(self.getMode() == _normParams._DatasetScope.GLOBAL)
        self._selectFileButton.setEnabled(
            self.getMode() == _normParams._DatasetScope.GLOBAL
        )
        if (
            self.getMode() == _normParams._DatasetScope.LOCAL
            and self.getGlobalFilePath() != self._FILE_PATH_LOCAL_VALUE
        ):
            self._lastGlobalPath = self.getGlobalFilePath()
            self.setGlobalFilePath(self._FILE_PATH_LOCAL_VALUE)
        elif (
            self.getMode() == _normParams._DatasetScope.GLOBAL
            and self.getGlobalFilePath() == self._FILE_PATH_LOCAL_VALUE
        ):
            if self._lastGlobalPath is not None:
                self.setGlobalFilePath(self._lastGlobalPath)

    def _selectFile(self):  # pragma: no cover
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(["HDF5 file *.h5 *.hdf5 *.nx *.nxs *.nexus"])

        if not dialog.exec_():
            dialog.close()
            return

        filesSelected = dialog.selectedFiles()
        if len(filesSelected) > 0:
            self.setGlobalFilePath(filesSelected[0])

    def _selectDataPath(self):
        """Open a dialog. If from a master file try to open the scan
        master file if any."""
        if self.getMode() is _normParams._DatasetScope.LOCAL:
            if self.getScan() is not None:
                scan = self.getScan()
                if not isinstance(scan, NXtomoScan):
                    mess = qt.QMessageBox(
                        parent=self,
                        icon=qt.QMessageBox.Warning,
                        text="local mode is only available for HDF5 acquisitions",
                    )
                    mess.setModal(False)
                    mess.show()
                    return
                else:
                    file_ = scan.master_file
            else:
                mess = qt.QMessageBox(
                    parent=self,
                    icon=qt.QMessageBox.Information,
                    text="No scan set. Unable to find the master file",
                )
                mess.setModal(False)
                mess.show()
                return
        elif self.getMode() is _normParams._DatasetScope.GLOBAL:
            file_ = self.getGlobalFilePath()
        else:
            raise ValueError(f"{self.getMode()} is not handled")

        dialog = DataFileDialog()
        dialog.selectFile(file_)
        dialog.setFilterMode(DataFileDialog.FilterMode.ExistingDataset)

        if not dialog.exec_():
            dialog.close()
            return
        else:
            selected_url = dialog.selectedUrl()
            self.setDataPath(DataUrl(path=selected_url).data_path())



from tomwer.utils import Launcher


def test_launcher():
    launcher = Launcher(prog="tomwer", version="1.0")
    launcher.add_command(
        "canvas", module_name="tomwer.app.canvas", description="open the orange-canvas"
    )
    launcher.print_help()
    launcher.execute_help(["tomwer", "canvas"])



# coding: utf-8

from tomwer.core.process.visualization.diffviewer import _DiffViewerPlaceHolder


def test_diff_viewer():
    process = _DiffViewerPlaceHolder(
        inputs={
            "data": None,
        }
    )
    process.run()



from tomwer.tests.conftest import qtapp  # noqa F401
import tempfile

import pytest
from silx.gui import qt

from tomwer.core.utils.scanutils import MockEDF
from tomwer.gui.control.datalist import GenericScanListDialog, VolumeList
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_DataListTest(
    qtapp,  # noqa F811
    tmp_path,
):
    widget = GenericScanListDialog(parent=None)
    widget._callbackRemoveAllFolders()
    widget.clear()
    folders = []
    for _ in range(5):
        folders.append(tempfile.mkdtemp())
        MockEDF.mockScan(scanID=folders[-1], nRadio=5, nRecons=5, nPagRecons=0, dim=10)

    for _folder in folders:
        widget.add(_folder)

    assert widget.n_scan() == len(folders)
    widget.remove(folders[0])
    assert widget.n_scan() == (len(folders) - 1)
    assert folders[0] not in widget.datalist._myitems
    widget.selectAll()
    widget._removeSelected()
    assert widget.n_scan() == 0

    widget.setAttribute(qt.Qt.WA_DeleteOnClose)
    widget.close()


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_VolumeList(
    qtapp,  # noqa F811
    tmp_path,
):
    """Test VolumeList behave as expected"""

    VolumeList(parent=None)



"""
environ
=======

This module contains some basic configuration options for Orange
(for now mostly changing directories where settings and data are saved).

How it works
------------

The configuration is read from '{sys.prefix}/etc/orangerc.cfg'
which is a standard `configparser` file.

orangerc.cfg
------------

.. code-block:: cfg

    # An exemple orangerc.cfg file
    # ----------------------------
    #
    # A number of variables are predefined:
    # - prefix: `sys.prefix`
    # - name: The application/library name ('Orange')
    # - version: The application/library name ('Orange.__version__')
    # - version.major, version.minor, version.micro: The version components

    [paths]
    # The base path where persistent application data can be stored
    # (here we define a prefix relative path)
    data_dir_base = %(prefix)s/share
    # the base path where application data can be stored
    cache_dir = %(prefix)s/cache/%(name)s/%(version)s

    # The following is only applicable for a running orange canvas application.

    # The base dir where widgets store their settings
    widget_settings_dir = %(prefix)s/config/%(name)s/widgets
    # The base dir where canvas stores its settings
    canvas_settings_dir = %(prefix)s/config/%(name)s/canvas

"""

from __future__ import annotations

import configparser
import os
import sys
import sysconfig

import tomwer


def _get_parsed_config():
    version = tomwer.__version__.split(".")
    data = sysconfig.get_path("data")
    vars = {
        "home": os.path.expanduser("~/"),
        "prefix": sys.prefix,
        "data": sysconfig.get_path("data"),
        "name": "Orange",
        "version": tomwer.__version__,
        "version.major": version[0],
        "version.minor": version[1],
        "version.micro": version[2],
    }
    conf = configparser.ConfigParser(vars)
    conf.read(
        [
            os.path.join(data, "etc/tomwerrc.conf"),
        ],
        encoding="utf-8",
    )
    if not conf.has_section("paths"):
        conf.add_section("paths")
    return conf


def get_path(name: str, default: str | None = None) -> str | None:
    """
    Get configured path

    :param name: The named config path value
    :param default: The default to return if `name` is not defined
    """
    cfg = _get_parsed_config()
    try:
        return cfg.get("paths", name)
    except (configparser.NoOptionError, configparser.NoSectionError):
        return default



# coding: utf-8
from __future__ import annotations


from silx.gui import qt

from tomwer.core.process.control.scantransfer import (
    ScanTransferTask as FolderTransfertP,
)
from tomwer.core.scan.scanbase import TomwerScanBase


class ScanTransfer(qt.QObject, FolderTransfertP):
    scanready = qt.Signal(TomwerScanBase)
    """emit when scan ready"""

    def __init__(self, parent=None, inputs=None, varinfo=None):
        qt.QObject.__init__(self, parent)
        FolderTransfertP.__init__(self, varinfo=varinfo, inputs=inputs)



"""
contains gui relative to semi-automatic axis calculation
"""

from __future__ import annotations


import logging

import numpy
from silx.gui import qt

from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.core.process.reconstruction.nabu.utils import _NabuMode, slice_index_to_int
from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisTask
from tomwer.gui import icons
from tomwer.gui.reconstruction.axis.CalculationWidget import CalculationWidget
from tomwer.gui.reconstruction.nabu.slices import NabuWidget
from tomwer.gui.reconstruction.nabu.platform import NabuPlatformSettings
from tomwer.gui.reconstruction.saaxis.corrangeselector import SliceAndCorWidget
from tomwer.gui.reconstruction.scores.control import ControlWidget
from tomwer.gui.reconstruction.scores.scoreplot import CorSelection
from tomwer.gui.reconstruction.scores.scoreplot import ScorePlot as _ScorePlot
from tomwer.gui.utils.buttons import TabBrowsersButtons
from tomwer.gui.utils.scandescription import ScanNameLabelAndShape
from tomwer.gui.settings import TAB_LABEL_PLATFORM_SETTINGS
from tomwer.synctools.axis import QAxisRP

_logger = logging.getLogger(__file__)


class ScorePlot(_ScorePlot, constructor=CorSelection):
    """Score plot dedicated to center of rotation.
    Redefine the current score value to display both the absolute
    and the relative values
    """

    def _updateScores(self):
        scan = self.__scan() if self.__scan else None
        img_width = None
        if scan is not None:
            if scan.saaxis_params:
                scan.saaxis_params.score_method = self.getScoreMethod()
                img_width = scan.dim_1
                # update autofocus
                SAAxisTask.autofocus(scan)

        self.setVarScores(
            scores=self._scores,
            score_method=self.getScoreMethod(),
            img_width=img_width,
            update_only_scores=True,
        )

    def _applyAutofocus(self):
        scan = self.__scan() if self.__scan else None
        if scan is None:
            return
        if scan.saaxis_params:
            best_cor = scan.saaxis_params.autofocus
            if best_cor:
                self._varSlider.setVarValue(best_cor)


class NabuAutoCorDiag(qt.QDialog):
    """
    GUI to compute an estimation of the Center Of Rotation
    """

    class CalculationWidget(CalculationWidget):
        def _modeChanged(self, *args, **kwargs):
            super()._modeChanged()
            self.getMethodLockPB().hide()

    sigRequestAutoCor = qt.Signal()
    """emit when user request auto cor"""

    def __init__(self, parent=None, qarixrp=None):
        assert qarixrp is not None, "An instance of QAxisRP should be provided"
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._automatic_cor = NabuAutoCorDiag.CalculationWidget(
            parent=self,
            axis_params=qarixrp,
        )
        self._automatic_cor.getMethodLockPB().hide()

        qcb = self._automatic_cor._qcbPosition
        for mode in (axis_mode.AxisMode.manual,):
            idx = qcb.findText(mode.value)
            if idx >= 0:
                qcb.removeItem(idx)

        self.layout().addWidget(self._automatic_cor)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(
            self.sigRequestAutoCor
        )
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("compute")


class _SAAxisTabWidget(qt.QTabWidget):
    sigConfigurationChanged = qt.Signal()
    """signal emitted each time the 'input' configuration changed.
    like slice to reconstruct, number of reconstruction, research width,
    nabu reconstruction parameters..."""

    def __init__(self, parent, backend=None):
        qt.QTabWidget.__init__(self, parent)
        # select slice & cor range

        self._sliceAndCorWidget = SliceAndCorWidget(self)
        sinogram_icon = icons.getQIcon("sinogram")
        self.addTab(self._sliceAndCorWidget, sinogram_icon, "slice && cor range")
        # nabu reconstruction parameters
        self._nabuSettings = NabuWidget(self)
        self._nabuSettings.setConfigurationLevel(level="required")
        self._nabuSettings.hideSlicesInterface()
        nabu_icon = icons.getQIcon("nabu")
        self.addTab(self._nabuSettings, nabu_icon, "reconstruction settings")
        # platform settings
        self._localPlatformSettings = NabuPlatformSettings(self)
        settings_icons = icons.getQIcon("parameters")
        self.addTab(
            self._localPlatformSettings, settings_icons, TAB_LABEL_PLATFORM_SETTINGS
        )
        # results
        self._resultsViewer = ScorePlot(self, variable_name="cor", backend=backend)
        results_icon = icons.getQIcon("results")
        self.addTab(self._resultsViewer, results_icon, "reconstructed slices")

        # connect signal / slot
        self._nabuSettings.sigConfigChanged.connect(self._configurationChanged)
        self._sliceAndCorWidget.sigConfigurationChanged.connect(
            self._configurationChanged
        )
        self._resultsViewer.sigConfigurationChanged.connect(self._configurationChanged)
        self.sigReconstructionSliceChanged = (
            self._sliceAndCorWidget.sigReconstructionSliceChanged
        )
        self.sigAutoCorRequested = self._sliceAndCorWidget.sigAutoCorRequested
        self.sigReconstructionRangeChanged = (
            self._sliceAndCorWidget.sigReconstructionRangeChanged
        )

        # expose function API
        self.setCorScores = self._resultsViewer.setVarScores
        self.setImgWidth = self._resultsViewer.setImgWidth
        self.setVoxelSize = self._resultsViewer.setVoxelSize
        self.setVolumeSize = self._resultsViewer.setVolumeSize
        self.setCurrentCorValue = self._resultsViewer.setCurrentVarValue
        self.getCurrentCorValue = self._resultsViewer.getCurrentVarValue
        self.getEstimatedCorPosition = self._sliceAndCorWidget.getEstimatedCorPosition
        self.setEstimatedCorPosition = self._sliceAndCorWidget.setEstimatedCorPosition
        self.getNReconstruction = self._sliceAndCorWidget.getNReconstruction
        self.setNReconstruction = self._sliceAndCorWidget.setNReconstruction
        self.getResearchWidth = self._sliceAndCorWidget.getResearchWidth
        self.setResearchWidth = self._sliceAndCorWidget.setResearchWidth
        self.getReconstructionSlices = self._sliceAndCorWidget.getReconstructionSlices
        self.setReconstructionSlices = self._sliceAndCorWidget.setReconstructionSlices
        self.getReconstructionMode = self._sliceAndCorWidget.getReconstructionMode
        self.setReconstructionMode = self._sliceAndCorWidget.setReconstructionMode
        self.getFrameWidth = self._sliceAndCorWidget.getFrameWidth
        self.setFrameWidth = self._sliceAndCorWidget.setFrameWidth
        self.setNabuReconsParams = self._nabuSettings.setConfiguration
        self.getNabuReconsParams = self._nabuSettings.getConfiguration
        self.getSlicesRange = self._sliceAndCorWidget.getSlicesRange
        self.setSlicesRange = self._sliceAndCorWidget.setSlicesRange
        self.loadSinogram = self._sliceAndCorWidget.loadSinogram
        self.saveReconstructedSlicesTo = self._resultsViewer.saveReconstructedSlicesTo
        # expose signals
        self.sigStartSinogramLoad = self._sliceAndCorWidget.sigStartSinogramLoad
        self.sigEndSinogramLoad = self._sliceAndCorWidget.sigEndSinogramLoad

    def showResults(self):
        self.setCurrentWidget(self._resultsViewer)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigurationChanged.emit()

    def lockAutoFocus(self, lock):
        self._resultsViewer.lockAutoFocus(lock=lock)

    def isAutoFocusLock(self):
        return self._resultsViewer.isAutoFocusLock()

    def hideAutoFocusButton(self):
        self._resultsViewer.hideAutoFocusButton()

    def getSinogramViewer(self):
        return self._sliceAndCorWidget._sinogramViewer

    def getCors(self, reference: str = "relative"):
        """Return cors to be computed"""
        if reference not in ("relative", "absolute"):
            raise ValueError("reference should be 'absolute' or 'relative'")
        return SAAxisParams.compute_cors(
            estimated_cor=self.getEstimatedCorPosition(reference),
            research_width=self.getResearchWidth(),
            n_reconstruction=self.getNReconstruction(),
        )

    def loadPreprocessingParams(self):
        """load reconstruction nabu if tomwer has already process this
        dataset. Not done for now"""
        return False

    def setScan(self, scan):
        self._resultsViewer.setScan(scan)
        self._nabuSettings.setScan(scan)
        if self.loadPreprocessingParams() and scan.axis_params is not None:
            self._nabuSettings.setConfiguration(scan.axis_params)
        self._sliceAndCorWidget.setScan(scan)

    def getConfiguration(self):
        nabu_config = self.getNabuReconsParams()
        enable_ht = int(self._nabuSettings.getMode() is _NabuMode.HALF_ACQ)
        nabu_config["reconstruction"]["enable_halftomo"] = enable_ht
        return {
            "research_width": self.getResearchWidth(),
            "n_reconstruction": self.getNReconstruction(),
            "slice_index": self.getReconstructionSlices(),
            "nabu_params": nabu_config,
            "mode": self.getReconstructionMode().value,
            "score_method": self.getScoreMethod().value,
            "estimated_cor": self.getEstimatedCorPosition(),
            "output_dir": self.getNabuReconsParams()
            .get("output", {})
            .get("location", None)
            or None,
        }

    def setConfiguration(self, config):
        if isinstance(config, SAAxisParams):
            config = config.to_dict()
        if not isinstance(config, dict):
            raise TypeError(
                f"config should be a dictionary or a SAAxisParams. Not {type(config)}"
            )

        research_width = config.get("research_width", None)
        if research_width is not None:
            self.setResearchWidth(research_width)
        n_reconstruction = config.get("n_reconstruction", None)
        if n_reconstruction is not None:
            self.setNReconstruction(n_reconstruction)
        estimated_cor = config.get("estimated_cor", None)
        if estimated_cor is not None:
            self.setEstimatedCorPosition(estimated_cor)
        slice_indexes = config.get("slice_index", None)
        if slice_indexes is not None:
            self.setReconstructionSlices(slice_indexes)
        if "nabu_params" in config:
            self.setNabuReconsParams(config["nabu_params"])
        if "mode" in config:
            self.setReconstructionMode(config["mode"])
        if "score_method" in config:
            self.setScoreMethod(config["score_method"])

    def getScoreMethod(self):
        return self._resultsViewer.getScoreMethod()

    def setScoreMethod(self, method):
        self._resultsViewer.setScoreMethod(method)

    def close(self):
        self._resultsViewer.close()
        self._resultsViewer = None
        self._sliceAndCorWidget.close()
        self._sliceAndCorWidget = None
        self._nabuSettings.close()
        self._nabuSettings = None
        super().close()


class SAAxisWindow(qt.QMainWindow):
    """
    Widget used to determine half-automatically the center of rotation
    """

    def __init__(self, parent=None, backend=None):
        qt.QMainWindow.__init__(self, parent)
        self._scan = None
        self._automaticCorWidget = None
        self._qaxis_rp = QAxisRP()
        self.setWindowFlags(qt.Qt.Widget)
        # central widget
        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QVBoxLayout())

        self._scanInfo = ScanNameLabelAndShape(self)
        self._mainWidget.layout().addWidget(self._scanInfo)
        self._tabWidget = _SAAxisTabWidget(self, backend=backend)
        self._mainWidget.layout().addWidget(self._tabWidget)
        self.setCentralWidget(self._mainWidget)
        # next and previous buttons for browsing the tab widget
        self._browserButtons = TabBrowsersButtons(self)
        self._dockWidgetBrwButtons = qt.QDockWidget(self)
        self._dockWidgetBrwButtons.setWidget(self._browserButtons)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetBrwButtons)
        self._dockWidgetBrwButtons.setFeatures(qt.QDockWidget.DockWidgetMovable)
        # control widget (validate, compute, cor positions)
        self._saaxisControl = ControlWidget(self)
        self._dockWidgetCtrl = qt.QDockWidget(self)
        self._dockWidgetCtrl.setWidget(self._saaxisControl)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetCtrl)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)

        # expose API
        self.setCorScores = self._tabWidget.setCorScores
        self.setImgWidth = self._tabWidget.setImgWidth
        self.setVoxelSize = self._tabWidget.setVoxelSize
        self.setVolumeSize = self._tabWidget.setVolumeSize
        self.setCurrentCorValue = self._tabWidget.setCurrentCorValue
        self.getCurrentCorValue = self._tabWidget.getCurrentCorValue
        self.getEstimatedCorPosition = self._tabWidget.getEstimatedCorPosition
        self.setEstimatedCorPosition = self._tabWidget.setEstimatedCorPosition
        self.getNReconstruction = self._tabWidget.getNReconstruction
        self.setNReconstruction = self._tabWidget.setNReconstruction
        self.getResearchWidth = self._tabWidget.getResearchWidth
        self.setResearchWidth = self._tabWidget.setResearchWidth
        self.getReconstructionSlices = self._tabWidget.getReconstructionSlices
        self.setReconstructionSlices = self._tabWidget.setReconstructionSlices
        self.getNabuReconsParams = self._tabWidget.getNabuReconsParams
        self.setNabuReconsParams = self._tabWidget.setNabuReconsParams
        self.getSlicesRange = self._tabWidget.getSlicesRange
        self.setSlicesRange = self._tabWidget.setSlicesRange
        self.getCors = self._tabWidget.getCors
        self.getMode = self._tabWidget.getReconstructionMode
        self.loadSinogram = self._tabWidget.loadSinogram
        self.saveReconstructedSlicesTo = self._tabWidget.saveReconstructedSlicesTo
        # expose signals
        self.sigValidated = self._saaxisControl.sigValidateRequest
        self.sigStartSinogramLoad = self._tabWidget.sigStartSinogramLoad
        self.sigEndSinogramLoad = self._tabWidget.sigEndSinogramLoad
        self.sigConfigurationChanged = self._tabWidget.sigConfigurationChanged

        # connect signal / slot
        self._tabWidget.sigReconstructionSliceChanged.connect(self._updateSinogramLine)
        self._tabWidget.sigAutoCorRequested.connect(self._autoCorRequested)
        self._tabWidget.sigReconstructionRangeChanged.connect(
            self._estimatedCorValueChanged
        )
        self._browserButtons.sigNextReleased.connect(self._showNextPage)
        self._browserButtons.sigPreviousReleased.connect(self._showPreviousPage)
        self._saaxisControl.sigComputationRequest.connect(self._launchReconstructions)
        self._saaxisControl.sigValidateRequest.connect(self._validate)

    def showResults(self):
        self._tabWidget.showResults()

    def getAutomaticCorWindow(self):
        if self._automaticCorWidget is None:
            self._automaticCorWidget = NabuAutoCorDiag(self, qarixrp=self._qaxis_rp)
            self._automaticCorWidget.setWindowTitle(
                "compute estimated center of rotation"
            )
            auto_cor_icon = icons.getQIcon("a")
            self._automaticCorWidget.setWindowIcon(auto_cor_icon)
            self._automaticCorWidget.sigRequestAutoCor.connect(
                self._computeEstimatedCor
            )
        return self._automaticCorWidget

    def compute(self):
        """force compute of the current scan"""
        self._saaxisControl.sigComputationRequest.emit()

    def getConfiguration(self) -> dict:
        return self._tabWidget.getConfiguration()

    def setConfiguration(self, config: dict):
        self._tabWidget.setConfiguration(config)

    def getQAxisRP(self):
        return self._qaxis_rp

    def setScan(self, scan):
        self._scan = scan
        self._tabWidget.setScan(scan)
        self._scanInfo.setScan(scan)
        self._updateSinogramLine()
        self._loadEstimatedCorFromScan(scan)

    def _loadEstimatedCorFromScan(self, scan):
        if scan.axis_params is not None:
            relative_cor = scan.axis_params.relative_cor_value
        else:
            relative_cor = None
        if relative_cor is None:
            relative_cor = scan.x_rotation_axis_pixel_position

        if relative_cor is not None and numpy.issubdtype(
            type(relative_cor), numpy.number
        ):
            self.setEstimatedCorPosition(relative_cor)

    def getScan(self):
        return self._scan

    def getScoreMethod(self):
        return self._tabWidget.getScoreMethod()

    def lockAutofocus(self, lock):
        self._tabWidget.lockAutoFocus(lock=lock)

    def isAutoFocusLock(self):
        return self._tabWidget.isAutoFocusLock()

    def hideAutoFocusButton(self):
        return self._tabWidget.hideAutoFocusButton()

    def _updateSinogramLine(self):
        r_slice = self.getReconstructionSlices()
        if r_slice == "middle":
            line = slice_index_to_int(slice_index="middle", scan=self._scan, axis="XY")
        else:
            line = list(r_slice.values())[0]
        self._tabWidget.getSinogramViewer().setLine(line)

    def _autoCorRequested(self):
        window = self.getAutomaticCorWindow()
        window.activateWindow()
        window.raise_()
        window.show()

    def _computeEstimatedCor(self) -> float | None:
        """callback when calculation of a estimated cor is requested.
        Should be implemted by OrangeWidget or application"""
        raise NotImplementedError("Base class")

    def _launchReconstructions(self):
        """callback when we want to launch the reconstruction of the
        slice for n cor value"""
        raise NotImplementedError("Base class")

    def _validate(self):
        raise NotImplementedError("Base class")

    def _estimatedCorValueChanged(self):
        cors = self.getCors("absolute")
        sino_viewer = self._tabWidget._sliceAndCorWidget._sinogramViewer
        estimated_cor = self.getEstimatedCorPosition("absolute")
        if estimated_cor == "middle":
            estimated_cor = 0

        if len(cors) < 2:
            return
        elif len(cors) > 2:
            other_cors = cors[1:-1]
        else:
            other_cors = ()
        sino_viewer.setCorRange(
            cor_min=cors[0],
            cor_max=cors[-1],
            estimated_cor=estimated_cor,
            other_cors=other_cors,
        )

    def _showNextPage(self, *args, **kwargs):
        idx = self._tabWidget.currentIndex()
        idx += 1
        if idx < self._tabWidget.count():
            self._tabWidget.setCurrentIndex(idx)

    def _showPreviousPage(self, *args, **kwargs):
        idx = self._tabWidget.currentIndex()
        idx -= 1
        if idx >= 0:
            self._tabWidget.setCurrentIndex(idx)

    def close(self):
        self._tabWidget.close()
        self._tabWidget = None
        super().close()


def _get_circle_plot(center, radius, n_pts=1000):
    pts = numpy.linspace(-numpy.pi, numpy.pi, n_pts)
    x = radius * numpy.cos(pts) + center[0]
    y = radius * numpy.sin(pts) + center[1]
    return x, y



from __future__ import annotations

import weakref
from silx.gui import qt
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThreadPerRun
from ewoksorange.gui.orange_imports import Input
from orangewidget import gui
from orangewidget.settings import Setting
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing

from tomwer.core.settings import TOMO_BEAMLINES
import tomwer.core.process.drac.publish
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.dataportal.publish import PublishProcessedDataWindow, has_ewoksnotify

from processview.core.superviseprocess import SuperviseProcess


class PublishProcessedDataOW(
    OWEwoksWidgetOneThreadPerRun,
    SuperviseProcess,
    WidgetLongProcessing,
    ewokstaskclass=tomwer.core.process.drac.publish.PublishICatDatasetTask,
):
    """
    This widget can receive 'data' (scan) and but some screenshot to be pushed on GALLERY.
    """

    name = "Data Portal Publishing"
    id = "orangecontrib.widgets.tomwer.dataportal.PublishProcessedDataOW.PublishProcessedDataOW"
    description = "Publish processed data to the (ESRF) data portal. \n To execute this you need to be inside the ESRF network"
    icon = "icons/publish_processed_data.svg"
    priority = 64
    keywords = [
        "tomography",
        "tomwer",
        "tomo_obj",
        "processed data",
        "PROCESSED_DATA",
        "publish",
        "icat",
        "icatplus",
        "pyicatplus",
        "pyicat-plus",
        "drac",
        "data portal",
        "dataportal",
        "portal",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _ewoks_default_inputs = Setting({})

    _ewoks_inputs_to_hide_from_orange = (
        "__process__",
        "beamline",
        "proposal",
        "dataset",
        "dry_run",
        "path",
    )

    class Inputs:
        # redefine the input to allow multiple and default
        data_portal_processed_datasets = Input(
            name="data_portal_processed_datasets",
            type=tuple,
            doc="data portal processed data to be saved",
            multiple=True,
            default=True,
        )

    def __init__(self, parent=None):
        super().__init__(parent)
        SuperviseProcess.__init__(self)
        WidgetLongProcessing.__init__(self)
        if not has_ewoksnotify:
            raise RuntimeError(
                "ewoksnotify not installed. Cannot define the PublishProcessedDataOW"
            )
        self.__on_going_task: int = 0
        """used to know how many tasks are on-going (as they are processed in parallel). Used to display processing wheel and processing"""

        self._window = PublishProcessedDataWindow(
            parent=self,
            beamlines=TOMO_BEAMLINES,
        )
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self._window)
        self._scan = None

        # load settings
        self._window.setConfiguration(self._ewoks_default_inputs)

        # connect signal / slot
        self._window.centralWidget().sigConfigChanged.connect(self._updateSettings)

    def _updateSettings(self):
        self._ewoks_default_inputs = self._window.getConfiguration()

    def setScan(self, scan: TomwerScanBase):
        self._scan = weakref.ref(scan)
        self._window.setScan(scan=scan)

    def getScan(self) -> TomwerScanBase | None:
        if self._scan is None:
            return None
        else:
            return self._scan()

    def _getScanFromIcatProcessedData(self) -> TomwerScanBase:
        """Return the associated scan to the latest set 'data_portal_processed_datasets'"""
        data_portal_processed_datasets = self.get_task_input_value(
            "data_portal_processed_datasets", None
        )
        if (
            data_portal_processed_datasets is None
            or len(data_portal_processed_datasets) == 0
        ):
            return None
        else:
            return data_portal_processed_datasets[0].source_scan

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.
        """
        # update the widget when receive the scan (proposal, dataset...)
        scan = self._getScanFromIcatProcessedData()
        # update processing wheel if necessary
        if self.__on_going_task == 0:
            self._startProcessing()
        self.__on_going_task += 1

        if scan is None:
            return

        if scan != self.getScan():
            self._window.setScan(scan)
        super().handleNewSignals()

    @Inputs.data_portal_processed_datasets
    def add(self, data_portal_processed_datasets, signal_id=None):
        # required because today ewoksorange is not handling multiple inputs
        self.set_dynamic_input(
            "data_portal_processed_datasets", data_portal_processed_datasets
        )

    def task_output_changed(self):
        # update processing wheel if necessary
        # task_output_changed is called when the ewoks task is finished from __ewoks_task_finished
        # call once even if there is 0..n outputs
        self.__on_going_task -= 1
        if self.__on_going_task == 0:
            self._endProcessing()
        return super().task_output_changed()

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()
        configuration = self._window.getConfiguration()

        task_inputs["beamline"] = configuration["beamline"]
        task_inputs["proposal"] = configuration["proposal"]
        task_inputs["__process__"] = weakref.ref(self)

        return task_inputs

    def sizeHint(self):
        return qt.QSize(500, 200)



import logging
from orangewidget import gui

import tomwer.core.process.edit.imagekeyeditor
from tomwer.gui.edit.imagekeyeditor import ImageKeyDialog
from ewokscore.missing_data import MissingData

from ...orange.managedprocess import TomwerWithStackStack

_logger = logging.getLogger(__name__)


class ImageKeyEditorOW(
    TomwerWithStackStack,
    ewokstaskclass=tomwer.core.process.edit.imagekeyeditor.ImageKeyEditorTask,
):
    """
    Widget to define on the fly the image_key of a NXtomoScan
    """

    name = "image-key-editor"
    id = "orange.widgets.tomwer.control.ImageKeyEditorOW.ImageKeyEditorOW"
    description = "Interface to edit `image_key` of nexus files"
    icon = "icons/image_key_editor.svg"
    priority = 24
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "edition",
        "NXTomo",
        "editor",
        "image key editor",
        "image-key-editor",
        "image_key",
        "image_key_control",
    ]

    want_main_area = True
    resizing_enabled = True

    _ewoks_inputs_to_hide_from_orange = ("configuration", "serialize_output_data")

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._scan = None
        layout = gui.vBox(self.mainArea, self.name).layout()

        self.widget = ImageKeyDialog(parent=self)
        layout.addWidget(self.widget)

        # connect signal / slot
        self.widget.sigValidated.connect(self._validateCallback)
        self.task_executor_queue.sigComputationStarted.connect(self._startProcessing)
        self.task_executor_queue.sigComputationEnded.connect(self._endProcessing)
        if isinstance(self.task_output_changed_callbacks, set):
            self.task_output_changed_callbacks.add(self._notify_state)
        elif isinstance(self.task_output_changed_callbacks, list):
            self.task_output_changed_callbacks.append(self._notify_state)
        else:
            raise NotImplementedError

    def handleNewSignals(self) -> None:
        scan = self.get_task_input_value("data", MissingData)
        if scan not in (None, MissingData):
            if scan is not self._scan:
                self._scan = scan
                self.widget.setScan(scan)
                self.notify_pending(scan=scan)
            self.activateWindow()
            self.raise_()
            self.show()

        # warning: avoid calling "handleNewSignals" else this will trigger the processing.
        # here we want for the user to explicitly click on 'validate' to modify the image keys

    def reprocess(self, scan):
        self.set_default_input(
            "data",
            scan,
        )

    def getConfiguration(self):
        return {
            "modifications": self.widget.getModifications(),
        }

    def _validateCallback(self):
        self.update_default_inputs(configuration=self.getConfiguration())
        self.update_default_inputs(
            serialize_output_data=False
        )  # avoid serializing when using orange
        self.execute_ewoks_task()

    def _notify_state(self):
        try:
            task_executor = self.sender()
            task_suceeded = task_executor.succeeded
            scan = task_executor.current_task.inputs.data
            if task_suceeded:
                self.notify_succeed(scan=scan)
            else:
                self.notify_failed(scan=scan)
        except Exception as e:
            _logger.error(f"failed to handle task finished callback. Reason is {e}")



# coding: utf-8

from __future__ import annotations

import logging

from silx.gui import qt

from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams

logger = logging.getLogger(__name__)


class QSAAxisParams(SAAxisParams, qt.QObject):
    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    sigAxisUrlChanged = qt.Signal()
    """Signal emitted when the axis url change"""

    def __init__(self):
        qt.QObject.__init__(self)
        SAAxisParams.__init__(self)

    def changed(self):
        self.sigChanged.emit()

    def axis_urls_changed(self):
        self.sigAxisUrlChanged.emit()
